import logging
import re
from os import environ
from unittest.mock import MagicMock, PropertyMock

import pytest
from qbittorrentapi import APINames, Client, exceptions
from qbittorrentapi._version_support import v
from qbittorrentapi.definitions import Dictionary, List
from qbittorrentapi.request import Request
from qbittorrentapi.torrents import TorrentDictionary, TorrentInfoList
from requests import Response

from tests.conftest import IS_QBT_DEV
from tests.utils import mkpath


def test_method_name(client, app_version):
    assert app_version == client._get("app", "version", response_class=str)
    assert app_version == client._get(
        APINames.Application,
        "version",
        response_class=str,
    )


def test_log_in():
    client_good = Client(VERIFY_WEBUI_CERTIFICATE=False)
    client_bad = Client(
        username="asdf",
        password="asdfasdf",
        VERIFY_WEBUI_CERTIFICATE=False,
    )

    client_good.auth_log_out()
    assert client_good.auth_log_in() is None
    assert client_good.is_logged_in is True
    client_good.auth_log_out()
    assert client_good.auth.log_in() is None
    assert client_good.auth.is_logged_in is True
    assert client_good.auth.is_logged_in is True
    with pytest.raises(exceptions.LoginFailed):
        client_bad.auth_log_in()
    with pytest.raises(exceptions.LoginFailed):
        client_bad.auth.log_in()


def test_log_in_via_auth():
    client_good = Client(VERIFY_WEBUI_CERTIFICATE=False)
    client_bad = Client(
        username="asdf",
        password="asdfasdf",
        VERIFY_WEBUI_CERTIFICATE=False,
    )

    assert (
        client_good.auth_log_in(
            username=environ.get("QBITTORRENTAPI_USERNAME"),
            password=environ.get("QBITTORRENTAPI_PASSWORD"),
        )
        is None
    )
    with pytest.raises(exceptions.LoginFailed):
        client_bad.auth_log_in(username="asdf", password="asdfasdf")


@pytest.mark.parametrize(
    "hostname",
    (
        "localhost:8080",
        "localhost:8080/",
        "http://localhost:8080",
        "http://localhost:8080/",
        "https://localhost:8080",
        "https://localhost:8080/",
        "//localhost:8080",
        "//localhost:8080/",
    ),
)
def test_hostname_format(app_version, hostname):
    client = Client(
        host=hostname,
        VERIFY_WEBUI_CERTIFICATE=False,
        REQUESTS_ARGS={"timeout": 1},
    )
    assert client.app.version == app_version
    # ensure the base URL is always normalized
    assert re.match(r"(http|https)://localhost:8080/", client._url._base_url)


@pytest.mark.parametrize(
    "hostname",
    (
        "localhost:8080/qbt",
        "localhost:8080/qbt/",
    ),
)
def test_hostname_user_base_path(hostname):
    client = Client(host=hostname, VERIFY_WEBUI_CERTIFICATE=False)
    # the command will fail but the URL will be built
    with pytest.raises(exceptions.APIConnectionError):
        _ = client.app.version
    # ensure user provided base paths are preserved
    assert re.match(r"(http|https)://localhost:8080/qbt/", client._url._base_url)


def test_port_from_host(app_version):
    host, port = environ.get("QBITTORRENTAPI_HOST").split(":")
    client = Client(host=host, port=port, VERIFY_WEBUI_CERTIFICATE=False)
    assert client.app.version == app_version


def _enable_disable_https(client, use_https):
    if use_https:
        client.app.preferences = {
            "use_https": True,
            "web_ui_https_cert_path": mkpath("/tmp", "_resources", "server.crt"),
            "web_ui_https_key_path": mkpath("/tmp", "_resources", "server.key"),
        }
    else:
        client.app.preferences = {"use_https": False}


@pytest.mark.skipif_before_api_version("2.2.1")
@pytest.mark.parametrize("use_https", (True, False))
def test_force_user_scheme(client, app_version, use_https):
    default_host = environ["QBITTORRENTAPI_HOST"]

    _enable_disable_https(client, use_https)

    client = Client(
        host="http://" + default_host,
        VERIFY_WEBUI_CERTIFICATE=False,
        FORCE_SCHEME_FROM_HOST=True,
        REQUESTS_ARGS={"timeout": 3},
    )
    if use_https:
        with pytest.raises(exceptions.APIConnectionError):
            assert client.app.version == app_version
    else:
        assert client.app.version == app_version
    assert client._url._base_url.startswith("http://")

    client = Client(
        host=default_host,
        VERIFY_WEBUI_CERTIFICATE=False,
        FORCE_SCHEME_FROM_HOST=True,
        REQUESTS_ARGS={"timeout": 3},
    )
    assert client.app.version == app_version
    if use_https:
        assert client._url._base_url.startswith("https://")
    else:
        assert client._url._base_url.startswith("http://")

    client = Client(
        host="https://" + default_host,
        VERIFY_WEBUI_CERTIFICATE=False,
        FORCE_SCHEME_FROM_HOST=True,
        REQUESTS_ARGS={"timeout": 3},
    )
    if not use_https:
        with pytest.raises(exceptions.APIConnectionError):
            assert client.app.version == app_version
    else:
        assert client.app.version == app_version
    assert client._url._base_url.startswith("https://")


@pytest.mark.skipif_before_api_version("2.2.1")
@pytest.mark.parametrize("scheme", ("http://", "https://"))
def test_both_https_http_not_working(client, app_version, scheme):
    default_host = environ["QBITTORRENTAPI_HOST"]
    _enable_disable_https(client, use_https=True)

    # rerun with verify=True
    test_client = Client(
        host=scheme + default_host,
        REQUESTS_ARGS={"timeout": 3},
    )
    with pytest.raises(exceptions.APIConnectionError):
        assert test_client.app.version == app_version
    assert test_client._url._base_url.startswith("https://")

    _enable_disable_https(client, use_https=False)


def test_legacy_env_vars():
    save_env_host = environ.get("QBITTORRENTAPI_HOST")
    save_env_username = environ.get("QBITTORRENTAPI_USERNAME")
    save_env_password = environ.get("QBITTORRENTAPI_PASSWORD")
    save_env_verify = environ.get("QBITTORRENTAPI_DO_NOT_VERIFY_WEBUI_CERTIFICATE")
    del environ["QBITTORRENTAPI_HOST"]
    del environ["QBITTORRENTAPI_USERNAME"]
    del environ["QBITTORRENTAPI_PASSWORD"]
    if "QBITTORRENTAPI_DO_NOT_VERIFY_WEBUI_CERTIFICATE" in environ:
        del environ["QBITTORRENTAPI_DO_NOT_VERIFY_WEBUI_CERTIFICATE"]

    client = Client()

    assert client.host == ""
    assert client.username == ""
    assert client._password == ""
    assert client._VERIFY_WEBUI_CERTIFICATE is True

    environ["PYTHON_QBITTORRENTAPI_HOST"] = excepted_host = "legacy:8090"
    environ["PYTHON_QBITTORRENTAPI_USERNAME"] = excepted_username = "legacyuser"
    environ["PYTHON_QBITTORRENTAPI_PASSWORD"] = expected_password = "legacypassword"
    environ["PYTHON_QBITTORRENTAPI_DO_NOT_VERIFY_WEBUI_CERTIFICATE"] = "true"

    client = Client()

    try:
        assert client.host == excepted_host
        assert client.username == excepted_username
        assert client._password == expected_password
        assert client._VERIFY_WEBUI_CERTIFICATE is False
    finally:
        environ["QBITTORRENTAPI_HOST"] = save_env_host
        environ["QBITTORRENTAPI_USERNAME"] = save_env_username
        environ["QBITTORRENTAPI_PASSWORD"] = save_env_password
        if save_env_verify is not None:
            environ["QBITTORRENTAPI_DO_NOT_VERIFY_WEBUI_CERTIFICATE"] = save_env_verify

    client = Client()

    assert client.host != excepted_host
    assert client.username != excepted_username
    assert client._password != expected_password
    assert client._VERIFY_WEBUI_CERTIFICATE is False

    del environ["PYTHON_QBITTORRENTAPI_HOST"]
    del environ["PYTHON_QBITTORRENTAPI_USERNAME"]
    del environ["PYTHON_QBITTORRENTAPI_PASSWORD"]
    del environ["PYTHON_QBITTORRENTAPI_DO_NOT_VERIFY_WEBUI_CERTIFICATE"]


def test_log_out(client):
    client.auth_log_out()
    with pytest.raises(exceptions.Forbidden403Error):
        # cannot call client.app.version directly since it will auto log back in
        client._request_manager("get", APINames.Application, "version")
    client.auth_log_in()
    client.auth.log_out()
    with pytest.raises(exceptions.Forbidden403Error):
        # cannot call client.app.version directly since it will auto log back in
        client._request_manager("get", APINames.Application, "version")
    client.auth_log_in()


def test_port(app_version):
    client = Client(host="localhost", port=8080, VERIFY_WEBUI_CERTIFICATE=False)
    assert client.app.version == app_version

    client = Client(host="localhost:8080", port=8081, VERIFY_WEBUI_CERTIFICATE=False)
    assert client.app.version == app_version


def test_response_str(client):
    response = MagicMock(spec_set=Response)

    type(response).text = PropertyMock(return_value="text response")
    assert client._cast(response, str) == "text response"

    type(response).text = PropertyMock(return_value="text response")
    with pytest.raises(exceptions.APIError, match="Exception during response parsing."):
        client._cast(response, int)


def test_response_int(client):
    response = MagicMock(spec_set=Response)

    type(response).text = PropertyMock(return_value="123")
    assert client._cast(response, int) == 123

    type(response).text = PropertyMock(return_value="text response")
    with pytest.raises(exceptions.APIError, match="Exception during response parsing."):
        client._cast(response, int)


def test_response_bytes(client):
    response = MagicMock(spec_set=Response)

    type(response).content = PropertyMock(return_value=b"bytes response")
    assert client._cast(response, bytes) == b"bytes response"


def test_response_json_list(client):
    response = MagicMock(spec_set=Response)

    response.json.return_value = ["json", "response"]
    assert client._cast(response, List) == ["json", "response"]

    response.json.return_value = 123
    with pytest.raises(exceptions.APIError, match="Exception during response parsing."):
        client._cast(response, List)

    del response.json

    type(response).text = PropertyMock(return_value='["json", "response"]')
    assert client._cast(response, List) == ["json", "response"]

    type(response).text = PropertyMock(return_value=123)
    with pytest.raises(exceptions.APIError, match="Exception during response parsing."):
        client._cast(response, List)


def test_response_json_dict(client):
    response = MagicMock(spec_set=Response)

    response.json.return_value = {"json": "response"}
    assert client._cast(response, Dictionary) == {"json": "response"}

    response.json.return_value = 123
    with pytest.raises(exceptions.APIError, match="Exception during response parsing."):
        client._cast(response, List)

    del response.json

    type(response).text = PropertyMock(return_value='{"json": "response"}')
    assert client._cast(response, Dictionary) == {"json": "response"}

    type(response).text = PropertyMock(return_value="123")
    with pytest.raises(exceptions.APIError, match="Exception during response parsing."):
        client._cast(response, List)


def test_response_unsupported(client):
    response = MagicMock(spec_set=Response)

    type(response).text = PropertyMock(return_value="123.01")
    with pytest.raises(
        exceptions.APIError,
        match="No handler defined to cast response.",
    ):
        client._cast(response, float)

    with pytest.raises(
        exceptions.APIError,
        match="No handler defined to cast response.",
    ):
        client._get(_name=APINames.Application, _method="version", response_class=float)


def test_simple_response(client):
    torrent = client.torrents_info()[0]
    assert isinstance(torrent, TorrentDictionary)
    torrent = client.torrents_info(SIMPLE_RESPONSE=True)[0]
    assert isinstance(torrent, dict)
    torrent = client.torrents_info(SIMPLE_RESPONSES=True)[0]
    assert isinstance(torrent, dict)
    torrent = client.torrents_info(SIMPLE_RESPONSE=False)[0]
    assert isinstance(torrent, TorrentDictionary)
    torrent = client.torrents_info(SIMPLE_RESPONSES=False)[0]
    assert isinstance(torrent, TorrentDictionary)
    client = Client(VERIFY_WEBUI_CERTIFICATE=False, SIMPLE_RESPONSES=True)
    torrent = client.torrents_info()[0]
    assert isinstance(torrent, dict)
    client = Client(VERIFY_WEBUI_CERTIFICATE=False, SIMPLE_RESPONSES=False)
    torrent = client.torrents_info()[0]
    assert isinstance(torrent, TorrentDictionary)


def test_request_extra_headers():
    client = Client(
        VERIFY_WEBUI_CERTIFICATE=False,
        EXTRA_HEADERS={"X-MY-HEADER": "asdf"},
    )
    client.auth.log_in()

    r = client._get(APINames.Application, "version")
    assert r.request.headers["X-MY-HEADER"] == "asdf"

    r = client._get(
        APINames.Application,
        "version",
        headers={"X-MY-HEADER-TWO": "zxcv"},
    )
    assert r.request.headers["X-MY-HEADER"] == "asdf"
    assert r.request.headers["X-MY-HEADER-TWO"] == "zxcv"

    r = client._get(
        APINames.Application,
        "version",
        headers={"X-MY-HEADER-TWO": "zxcv"},
        requests_args={"headers": {"X-MY-HEADER-THREE": "tyui"}},
    )
    assert r.request.headers["X-MY-HEADER"] == "asdf"
    assert r.request.headers["X-MY-HEADER-TWO"] == "zxcv"
    assert r.request.headers["X-MY-HEADER-THREE"] == "tyui"

    r = client._get(
        APINames.Application,
        "version",
        headers={"X-MY-HEADER": "zxcv"},
        requests_args={"headers": {"X-MY-HEADER": "tyui"}},
    )
    assert r.request.headers["X-MY-HEADER"] == "zxcv"

    r = client._get(APINames.Application, "version", headers={"X-MY-HEADER": "zxcv"})
    assert r.request.headers["X-MY-HEADER"] == "zxcv"


@pytest.mark.skipif_before_api_version("2.2.1")
def test_requests_timeout(api_version):
    class MyTimeoutError(Exception):
        pass

    logger = logging.getLogger("test_requests_timeout")

    timeout = 1e-100
    loops = 1000
    client = Client(VERIFY_WEBUI_CERTIFICATE=False)
    with pytest.raises(MyTimeoutError):
        try:
            for _ in range(loops):
                client.torrents_info(requests_args={"timeout": timeout})
        except exceptions.APIConnectionError as exp:
            logger.error("%r", exp)
            if "ReadTimeoutError" in str(exp) or "RemoteDisconnected" in str(exp):
                raise MyTimeoutError

    client = Client(VERIFY_WEBUI_CERTIFICATE=False, REQUESTS_ARGS={"timeout": timeout})
    with pytest.raises(MyTimeoutError):
        try:
            for _ in range(loops):
                client.torrents_info()
        except exceptions.APIConnectionError as exp:
            logger.error("%r", exp)
            if "ReadTimeoutError" in str(exp) or "RemoteDisconnected" in str(exp):
                raise MyTimeoutError


def test_request_extra_params(client, orig_torrent):
    """Extra params can be sent directly to qBittorrent but there aren't any real use-
    cases so force it."""
    response = client._post(
        APINames.Torrents,
        "info",
        hashes=orig_torrent.hash,
        extra_param="extra",
    )
    assert "extra_param=extra" in response.request.body
    torrent = TorrentInfoList(response.json(), client)[0]
    assert torrent.hash == orig_torrent.hash

    response = client._get(
        APINames.Torrents,
        "info",
        hashes=orig_torrent.hash,
        extra_param="extra",
    )
    assert "extra_param=extra" in response.request.url
    torrent = TorrentInfoList(response.json(), client)[0]
    assert torrent.hash == orig_torrent.hash


def test_unsupported_version_error(monkeypatch):
    if IS_QBT_DEV:
        return

    client = Client(
        VERIFY_WEBUI_CERTIFICATE=False,
        RAISE_ERROR_FOR_UNSUPPORTED_QBITTORRENT_VERSIONS=True,
    )
    monkeypatch.setattr(client, "app_version", MagicMock(return_value="1.0.0"))
    with pytest.raises(exceptions.UnsupportedQbittorrentVersion):
        client.app_web_api_version()

    client = Client(
        VERIFY_WEBUI_CERTIFICATE=False,
        RAISE_ERROR_FOR_UNSUPPORTED_QBITTORRENT_VERSIONS=True,
    )
    client.app_version()


def test_disable_logging():
    Client(DISABLE_LOGGING_DEBUG_OUTPUT=False)
    assert logging.getLogger("qbittorrentapi").level == logging.NOTSET
    assert logging.getLogger("requests").level == logging.NOTSET
    assert logging.getLogger("urllib3").level == logging.NOTSET

    Client(DISABLE_LOGGING_DEBUG_OUTPUT=True)
    assert logging.getLogger("qbittorrentapi").level == logging.INFO
    assert logging.getLogger("requests").level == logging.INFO
    assert logging.getLogger("urllib3").level == logging.INFO

    logging.getLogger("qbittorrentapi").setLevel(level=logging.CRITICAL)
    Client(DISABLE_LOGGING_DEBUG_OUTPUT=True)
    assert logging.getLogger("qbittorrentapi").level == logging.CRITICAL
    assert logging.getLogger("requests").level == logging.INFO
    assert logging.getLogger("urllib3").level == logging.INFO


def test_verify_cert(app_version):
    client = Client(VERIFY_WEBUI_CERTIFICATE=False)
    assert client._VERIFY_WEBUI_CERTIFICATE is False
    assert client.app.version == app_version

    # this is only ever going to work with a trusted cert....disabling for now
    # client = Client(VERIFY_WEBUI_CERTIFICATE=True)
    # assert client._VERIFY_WEBUI_CERTIFICATE is True
    # assert client.app.version == app_version

    environ["QBITTORRENTAPI_DO_NOT_VERIFY_WEBUI_CERTIFICATE"] = "true"
    client = Client(VERIFY_WEBUI_CERTIFICATE=True)
    assert client._VERIFY_WEBUI_CERTIFICATE is False
    assert client.app.version == app_version
    del environ["QBITTORRENTAPI_DO_NOT_VERIFY_WEBUI_CERTIFICATE"]


def test_api_connection_error():
    with pytest.raises(exceptions.APIConnectionError):
        Client(host="localhost:8081").auth_log_in(_retries=0)

    with pytest.raises(exceptions.APIConnectionError):
        Client(host="localhost:8081").auth_log_in(_retries=1)

    with pytest.raises(exceptions.APIConnectionError):
        Client(host="localhost:8081").auth_log_in(_retries=2)

    with pytest.raises(exceptions.APIConnectionError):
        Client(host="localhost:8081").auth_log_in(_retries=3)


def test_http400(client, app_version, orig_torrent):
    with pytest.raises(exceptions.MissingRequiredParameters400Error):
        client.torrents_file_priority(torrent_hash=orig_torrent.hash)

    if v(app_version) >= v("4.1.6"):
        with pytest.raises(exceptions.InvalidRequest400Error) as exc_info:
            client.torrents_file_priority(
                torrent_hash=orig_torrent.hash,
                file_ids="asdf",
                priority="asdf",
            )
        assert exc_info.value.http_status_code == 400


def test_http401():
    client = Client(VERIFY_WEBUI_CERTIFICATE=False)
    _ = client.app.preferences
    # ensure cross site scripting protection is enabled
    client.app.preferences = dict(web_ui_csrf_protection_enabled=True)
    # simulate a XSS request
    with pytest.raises(exceptions.Unauthorized401Error) as exc_info:
        client.app_version(
            headers={"Origin": "https://example.com", "Referer": "https://example.com"}
        )
    assert exc_info.value.http_status_code == 401


@pytest.mark.parametrize("params", ({}, {"hash": "asdf"}, {"hashes": "asdf|asdf"}))
def test_http404(client, params):
    with pytest.raises(exceptions.NotFound404Error, match="Torrent hash") as exc_info:
        client.torrents_rename(torrent_hash="zxcv", new_torrent_name="erty")
    assert "zxcv" in exc_info.value.args[0]

    response = MagicMock(spec=Response, status_code=404, text="", request=object())
    with pytest.raises(exceptions.HTTPError, match="") as exc_info:
        Request._handle_error_responses(data={}, params=params, response=response)
    assert exc_info.value.http_status_code == 404
    if params:
        assert params[list(params.keys())[0]] in exc_info.value.args[0]

    response = MagicMock(
        spec=Response, status_code=404, text="unexpected msg", request=object()
    )
    with pytest.raises(exceptions.HTTPError, match="unexpected msg") as exc_info:
        Request._handle_error_responses(data={}, params=params, response=response)
    assert exc_info.value.http_status_code == 404
    assert exc_info.value.args[0] == "unexpected msg"

    response = MagicMock(spec=Response, status_code=404, text="", request=object())
    with pytest.raises(exceptions.HTTPError, match="") as exc_info:
        Request._handle_error_responses(data=params, params={}, response=response)
    assert exc_info.value.http_status_code == 404
    if params:
        assert params[list(params.keys())[0]] in exc_info.value.args[0]

    response = MagicMock(
        spec=Response, status_code=404, text="unexpected msg", request=object()
    )
    with pytest.raises(exceptions.HTTPError, match="unexpected msg") as exc_info:
        Request._handle_error_responses(data=params, params={}, response=response)
    assert exc_info.value.http_status_code == 404
    assert exc_info.value.args[0] == "unexpected msg"


def test_http405(client, app_version):
    # v4.4.4 uses same API version as previous release...
    if v(app_version) >= v("v4.4.4"):
        with pytest.raises(exceptions.MethodNotAllowed405Error) as exc_info:
            client._get(APINames.Authorization, "logout")
        assert exc_info.value.http_status_code == 405


def test_http409(client, app_version):
    if v(app_version) >= v("4.1.6"):
        with pytest.raises(exceptions.Conflict409Error) as exc_info:
            client.torrents_set_location(torrent_hashes="asdf", location="/etc/asdf/")
        assert exc_info.value.http_status_code == 409


def test_http415(client):
    with pytest.raises(exceptions.UnsupportedMediaType415Error) as exc_info:
        client.torrents.add(torrent_files="/etc/hosts")
    assert exc_info.value.http_status_code == 415


@pytest.mark.parametrize("status_code", (500, 503))
def test_http500(status_code):
    response = MagicMock(
        spec=Response, status_code=status_code, text="asdf", request=object()
    )
    with pytest.raises(exceptions.InternalServerError500Error) as exc_info:
        Request._handle_error_responses(data={}, params={}, response=response)
    assert exc_info.value.http_status_code == status_code


@pytest.mark.parametrize("status_code", (402, 406))
def test_http_error(status_code):
    response = MagicMock(
        spec=Response, status_code=status_code, text="asdf", request=object()
    )
    with pytest.raises(exceptions.HTTPError) as exc_info:
        Request._handle_error_responses(data={}, params={}, response=response)
    assert exc_info.value.http_status_code == status_code


def test_request_retry_success(monkeypatch, caplog):
    def request500(*args, **kwargs):
        raise exceptions.HTTP500Error()

    client = Client(VERIFY_WEBUI_CERTIFICATE=False)
    client.auth_log_in()
    with monkeypatch.context() as m:
        m.setattr(client, "_request", request500)
        with caplog.at_level(logging.DEBUG, logger="qbittorrentapi"), pytest.raises(
            exceptions.HTTP500Error
        ):
            client.app_version()
        assert "Retry attempt" in caplog.text


def test_request_retry_skip(caplog):
    client = Client(VERIFY_WEBUI_CERTIFICATE=False)
    client.auth_log_in()
    with caplog.at_level(logging.DEBUG, logger="qbittorrentapi"), pytest.raises(
        exceptions.MissingRequiredParameters400Error
    ):
        client.torrents_rename()
    assert "Retry attempt" not in caplog.text


def test_verbose_logging(caplog):
    client = Client(VERBOSE_RESPONSE_LOGGING=True, VERIFY_WEBUI_CERTIFICATE=False)
    with caplog.at_level(logging.DEBUG, logger="qbittorrentapi"), pytest.raises(
        exceptions.NotFound404Error
    ):
        client.torrents_rename(torrent_hash="asdf", new_torrent_name="erty")
    assert "Response status" in caplog.text


def test_stack_printing(capsys):
    client = Client(VERIFY_WEBUI_CERTIFICATE=False)
    client._PRINT_STACK_FOR_EACH_REQUEST = True
    client.app_version()

    assert "print_stack()" in capsys.readouterr().err


def test_auto_authentication(caplog, app_version):
    client = Client(
        RAISE_NOTIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS=True,
        VERIFY_WEBUI_CERTIFICATE=False,
    )
    # check if first API call works if already logged in
    client.auth_log_in()
    with caplog.at_level(logging.DEBUG, logger="qbittorrentapi"):
        qbt_version = client.app.version
    assert "Login may have expired...attempting new login" not in caplog.text
    assert qbt_version == app_version

    # ensure login happens after first API call fails
    client.auth_log_out()
    with caplog.at_level(logging.DEBUG, logger="qbittorrentapi"):
        qbt_version = client.app.version
    assert "Login may have expired...attempting new login" in caplog.text
    assert qbt_version == app_version


def test_not_implemented_no_error(monkeypatch):
    client = Client(VERIFY_WEBUI_CERTIFICATE=False)

    monkeypatch.setattr(client, "app_web_api_version", MagicMock(return_value="1.0.0"))
    assert client.search_categories() is None

    monkeypatch.setattr(client, "app_web_api_version", MagicMock(return_value="10.0.0"))
    assert client.search_categories() is None


def test_not_implemented_error(monkeypatch, client):
    monkeypatch.setattr(client, "app_web_api_version", MagicMock(return_value="1.0.0"))
    with pytest.raises(NotImplementedError, match=r"This endpoint is available"):
        client.search_categories()

    monkeypatch.setattr(client, "app_web_api_version", MagicMock(return_value="10.0.0"))
    with pytest.raises(NotImplementedError, match=r"This endpoint was removed"):
        client.search_categories()



import sys

import pytest
from qbittorrentapi import APINames
from qbittorrentapi.sync import SyncMainDataDictionary, SyncTorrentPeersDictionary


@pytest.mark.skipif(sys.version_info < (3, 9), reason="removeprefix not in 3.8")
def test_methods(client):
    namespace = APINames.Sync
    all_dotted_methods = set(dir(getattr(client, namespace)))

    for meth in [meth for meth in dir(client) if meth.startswith(f"{namespace}_")]:
        assert meth.removeprefix(f"{namespace}_") in all_dotted_methods


@pytest.mark.parametrize("maindata_func", ["sync_maindata", "sync.maindata"])
@pytest.mark.parametrize("rid", [None, 0, 1, 100000])
def test_sync_maindata(client, maindata_func, rid):
    assert isinstance(client.func(maindata_func)(rid=rid), SyncMainDataDictionary)


@pytest.mark.parametrize("rid", (None, 0, 1, 100000))
def test_sync_maindata_delta(client, rid):
    assert isinstance(client.sync.maindata(rid=rid), SyncMainDataDictionary)
    assert isinstance(client.sync.maindata.delta(), SyncMainDataDictionary)
    assert isinstance(client.sync.maindata.delta(), SyncMainDataDictionary)


def test_sync_maindata_reset(client):
    client.sync.maindata.delta()
    assert client.sync.maindata._rid != 0
    client.sync.maindata.reset_rid()
    assert client.sync.maindata._rid == 0


@pytest.mark.parametrize("rid", (None, 0, 1, 100000))
def test_sync_torrent_peers(client, rid, orig_torrent):
    assert isinstance(
        client.sync_torrent_peers(rid=rid, torrent_hash=orig_torrent.hash),
        SyncTorrentPeersDictionary,
    )
    assert isinstance(
        client.sync.torrent_peers(rid=rid, torrent_hash=orig_torrent.hash),
        SyncTorrentPeersDictionary,
    )
    assert isinstance(
        client.sync.torrent_peers.delta(torrent_hash=orig_torrent.hash),
        SyncTorrentPeersDictionary,
    )

    client.sync.torrent_peers.reset_rid()
    assert client.sync.torrent_peers._rid == 0



from __future__ import annotations

from typing import Any, Mapping

from qbittorrentapi.log import LogAPIMixIn
from qbittorrentapi.rss import RSSAPIMixIn
from qbittorrentapi.search import SearchAPIMixIn
from qbittorrentapi.sync import SyncAPIMixIn
from qbittorrentapi.torrentcreator import TorrentCreatorAPIMixIn
from qbittorrentapi.torrents import TorrentsAPIMixIn
from qbittorrentapi.transfer import TransferAPIMixIn

# NOTES
# Implementation
#     Required API parameters
#         - To avoid runtime errors, required API parameters are not explicitly
#           enforced in the code. Instead, I found if qBittorrent returns HTTP400
#           without am error message, at least one required parameter is missing.
#           This raises a MissingRequiredParameters400 error.
#         - Alternatively, if a parameter is malformatted, HTTP400 is returned
#           with an error message.
#           This raises a InvalidRequest400 error.
#
#     Unauthorized HTTP 401
#         - This is only raised if XSS is detected or host header validation fails.
#
# API Peculiarities
#     app/setPreferences
#         - This was endlessly frustrating since it requires data in the
#           form of {'json': dumps({'dht': True})}...this way, Requests sends the
#           JSON dump as a key/value pair for "json" via x-www-form-urlencoded.
#         - Sending an empty string for 'banned_ips' drops the useless message
#           below in to the log file (same for WebUI):
#             ' is not a valid IP address and was rejected while applying the list of banned addresses.'  # noqa: E501
#             - [Resolved] https://github.com/qbittorrent/qBittorrent/issues/10745
#
#     torrents/downloadLimit and uploadLimit
#         - Hashes handling is non-standard. 404 is not returned for bad hashes and 'all' doesn't work.  # noqa: E501
#         - https://github.com/qbittorrent/qBittorrent/blob/6de02b0f2a79eeb4d7fb624c39a9f65ffe181d68/src/webui/api/torrentscontroller.cpp#L754  # noqa: E501
#         - https://github.com/qbittorrent/qBittorrent/issues/10744
#
#     torrents/info
#         - when using a GET request, the params (such as category) seemingly can't
#           contain spaces; however, POST does work with spaces.
#         - [Resolved] https://github.com/qbittorrent/qBittorrent/issues/10606


class Client(
    LogAPIMixIn,
    SyncAPIMixIn,
    TransferAPIMixIn,
    TorrentsAPIMixIn,
    TorrentCreatorAPIMixIn,
    RSSAPIMixIn,
    SearchAPIMixIn,
):
    """
    Initialize API for qBittorrent client.

    Host must be specified. Username and password can be specified at login.
    A call to :meth:`~qbittorrentapi.auth.AuthAPIMixIn.auth_log_in` is not explicitly
    required if username and password are provided during Client construction.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host='localhost:8080', username='admin', password='adminadmin')
        >>> torrents = client.torrents_info()

    :param host: hostname for qBittorrent Web API, ``[http[s]://]localhost[:8080][/path]``
    :param port: port number for qBittorrent Web API (ignored if host contains a port)
    :param username: username for qBittorrent Web API
    :param password: password for qBittorrent Web API
    :param SIMPLE_RESPONSES: By default, complex objects are returned from some
        endpoints. These objects will allow for accessing responses' items as
        attributes and include methods for contextually relevant actions. This
        comes at the cost of performance. Generally, this cost isn't large;
        however, some endpoints, such as ``torrents_files()`` method, may need
        to convert a large payload. Set this to True to return the simple JSON
        back. Alternatively, set this to True only for an individual method call.
        For instance, when requesting the files for a torrent:
        ``client.torrents_files(torrent_hash='...', SIMPLE_RESPONSES=True)``
    :param VERIFY_WEBUI_CERTIFICATE: Set to False to skip verify certificate for
        HTTPS connections; for instance, if the connection is using a self-signed
        certificate. Not setting this to False for self-signed certs will cause a
        :class:`~qbittorrentapi.exceptions.APIConnectionError` exception to be raised.
    :param EXTRA_HEADERS: Dictionary of HTTP Headers to include in all requests
        made to qBittorrent.
    :param REQUESTS_ARGS: Dictionary of configuration for Requests package:
        `<https://requests.readthedocs.io/en/latest/api/#requests.request>`_
    :param FORCE_SCHEME_FROM_HOST: If a scheme (i.e. ``http`` or ``https``) is
        specified in host, it will be used regardless of whether qBittorrent is
        configured for HTTP or HTTPS communication. Normally, this client will
        attempt to determine which scheme qBittorrent is actually listening on...
        but this can cause problems in rare cases. Defaults ``False``.
    :param RAISE_NOTIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS: Some Endpoints
        may not be implemented in older versions of qBittorrent. Setting this to True
        will raise a :class:`NotImplementedError` instead of just returning``None``.
    :param RAISE_ERROR_FOR_UNSUPPORTED_QBITTORRENT_VERSIONS: raise the
        UnsupportedQbittorrentVersion exception if the connected version of
        qBittorrent is not fully supported by this client. Defaults ``False``.
    :param DISABLE_LOGGING_DEBUG_OUTPUT: Turn off debug output from logging for
        this package as well as Requests & urllib3.
    """  # noqa: E501

    def __init__(
        self,
        host: str = "",
        port: str | int | None = None,
        username: str | None = None,
        password: str | None = None,
        EXTRA_HEADERS: Mapping[str, str] | None = None,
        REQUESTS_ARGS: Mapping[str, Any] | None = None,
        VERIFY_WEBUI_CERTIFICATE: bool = True,
        FORCE_SCHEME_FROM_HOST: bool = False,
        RAISE_NOTIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS: bool = False,
        RAISE_ERROR_FOR_UNSUPPORTED_QBITTORRENT_VERSIONS: bool = False,
        VERBOSE_RESPONSE_LOGGING: bool = False,
        SIMPLE_RESPONSES: bool = False,
        DISABLE_LOGGING_DEBUG_OUTPUT: bool = False,
    ):
        super().__init__(
            host=host,
            port=port,
            username=username,
            password=password,
            EXTRA_HEADERS=EXTRA_HEADERS,
            REQUESTS_ARGS=REQUESTS_ARGS,
            VERIFY_WEBUI_CERTIFICATE=VERIFY_WEBUI_CERTIFICATE,
            FORCE_SCHEME_FROM_HOST=FORCE_SCHEME_FROM_HOST,
            RAISE_NOTIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS=RAISE_NOTIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS,  # noqa: E501
            RAISE_ERROR_FOR_UNSUPPORTED_QBITTORRENT_VERSIONS=RAISE_ERROR_FOR_UNSUPPORTED_QBITTORRENT_VERSIONS,
            VERBOSE_RESPONSE_LOGGING=VERBOSE_RESPONSE_LOGGING,
            SIMPLE_RESPONSES=SIMPLE_RESPONSES,
            DISABLE_LOGGING_DEBUG_OUTPUT=DISABLE_LOGGING_DEBUG_OUTPUT,
        )



import errno
import platform
import sys
from time import sleep

import pytest
import requests
from qbittorrentapi import APINames
from qbittorrentapi._version_support import v
from qbittorrentapi.exceptions import (
    Conflict409Error,
    Forbidden403Error,
    InvalidRequest400Error,
    TorrentFileError,
    TorrentFileNotFoundError,
    TorrentFilePermissionError,
)
from qbittorrentapi.torrents import (
    TagList,
    TorrentCategoriesDictionary,
    TorrentFilesList,
    TorrentInfoList,
    TorrentLimitsDictionary,
    TorrentPieceInfoList,
    TorrentPropertiesDictionary,
    TorrentsAddPeersDictionary,
    TrackersList,
    WebSeedsList,
)

from tests.conftest import (
    ROOT_FOLDER_TORRENT_FILE,
    ROOT_FOLDER_TORRENT_HASH,
    TORRENT1_FILENAME,
    TORRENT1_HASH,
    TORRENT1_URL,
    TORRENT2_FILENAME,
    TORRENT2_HASH,
    TORRENT2_URL,
    new_torrent_standalone,
)
from tests.utils import check, mkpath, retry


def disable_queueing(client):
    if client.app.preferences.queueing_enabled:
        client.app.set_preferences(dict(queueing_enabled=False))


def enable_queueing(client):
    if not client.app.preferences.queueing_enabled:
        client.app.set_preferences(dict(queueing_enabled=True))


@pytest.mark.skipif(sys.version_info < (3, 9), reason="removeprefix not in 3.8")
def test_methods(client):
    all_dotted_methods = {
        meth
        for namespace in [APINames.Torrents, "torrent_tags", "torrent_categories"]
        for meth in dir(getattr(client, namespace))
    }

    for meth in [meth for meth in dir(client) if meth.startswith("torrents_")]:
        assert meth.removeprefix("torrents_") in all_dotted_methods


# something was wrong with torrents_add on v2.0.0 (the initial version)
@pytest.mark.skipif_before_api_version("2.0.1")
@pytest.mark.parametrize(
    "add_func, delete_func",
    [("torrents_add", "torrents_delete"), ("torrents.add", "torrents.delete")],
)
def test_add_delete(client, add_func, delete_func, tmp_path):
    def download_file(url, filename=None, return_bytes=False):
        max_attempts = 3
        for attempt in range(max_attempts):
            try:
                with requests.get(url, timeout=30) as r:
                    r.raise_for_status()
                    if return_bytes:
                        return r.content
                    with open(mkpath(tmp_path, filename), "wb") as f:
                        for chunk in r.iter_content(chunk_size=1024):
                            f.write(chunk)
            except Exception if attempt < (max_attempts - 1) else ZeroDivisionError:
                pass  # throw away errors until we hit the retry limit
            else:
                return
        raise Exception("Download failed: %s" % url)

    def delete():
        client.func(delete_func)(delete_files=True, torrent_hashes=TORRENT1_HASH)
        client.func(delete_func)(delete_files=True, torrent_hashes=TORRENT2_HASH)
        check(
            lambda: [t.hash for t in client.torrents_info()],
            TORRENT2_HASH,
            reverse=True,
            negate=True,
        )

    def check_torrents_added(f):
        def inner(**kwargs):
            try:
                f(**kwargs)
                check(
                    lambda: [t.hash for t in client.torrents_info()],
                    TORRENT1_HASH,
                    reverse=True,
                )
                if kwargs.get("single", False) is False:
                    check(
                        lambda: [t.hash for t in client.torrents_info()],
                        TORRENT2_HASH,
                        reverse=True,
                    )
            finally:
                delete()

        return inner

    @retry()
    @check_torrents_added
    def add_by_filename(single):
        download_file(url=TORRENT1_URL, filename=TORRENT1_FILENAME)
        download_file(url=TORRENT2_URL, filename=TORRENT2_FILENAME)
        files = (
            mkpath(tmp_path, TORRENT1_FILENAME),
            mkpath(tmp_path, TORRENT2_FILENAME),
        )

        if single:
            assert client.func(add_func)(torrent_files=files[0]) == "Ok."
        else:
            assert client.func(add_func)(torrent_files=files) == "Ok."

    @retry()
    @check_torrents_added
    def add_by_filename_dict(single):
        download_file(url=TORRENT1_URL, filename=TORRENT1_FILENAME)
        download_file(url=TORRENT2_URL, filename=TORRENT2_FILENAME)

        if single:
            assert (
                client.func(add_func)(
                    torrent_files={
                        TORRENT1_FILENAME: mkpath(tmp_path, TORRENT1_FILENAME)
                    }
                )
                == "Ok."
            )
        else:
            files = {
                TORRENT1_FILENAME: mkpath(tmp_path, TORRENT1_FILENAME),
                TORRENT2_FILENAME: mkpath(tmp_path, TORRENT2_FILENAME),
            }
            assert client.func(add_func)(torrent_files=files) == "Ok."

    @retry()
    @check_torrents_added
    def add_by_filehandles(single):
        download_file(url=TORRENT1_URL, filename=TORRENT1_FILENAME)
        download_file(url=TORRENT2_URL, filename=TORRENT2_FILENAME)
        files = (
            open(mkpath(tmp_path, TORRENT1_FILENAME), "rb"),  # noqa: SIM115
            open(mkpath(tmp_path, TORRENT2_FILENAME), "rb"),  # noqa: SIM115
        )

        if single:
            assert client.func(add_func)(torrent_files=files[0]) == "Ok."
        else:
            assert client.func(add_func)(torrent_files=files) == "Ok."

        for file in files:
            file.close()

    @retry()
    @check_torrents_added
    def add_by_bytes(single):
        files = (
            download_file(TORRENT1_URL, return_bytes=True),
            download_file(TORRENT2_URL, return_bytes=True),
        )

        if single:
            assert client.func(add_func)(torrent_files=files[0]) == "Ok."
        else:
            assert client.func(add_func)(torrent_files=files) == "Ok."

    @retry()
    @check_torrents_added
    def add_by_url(single):
        urls = (TORRENT1_URL, TORRENT2_URL)

        if single:
            client.func(add_func)(urls=urls[0])
        else:
            client.func(add_func)(urls=urls)

    add_by_filename(single=False)
    add_by_filename(single=True)
    add_by_filename_dict(single=False)
    add_by_filename_dict(single=True)
    add_by_url(single=False)
    add_by_url(single=True)
    add_by_filehandles(single=False)
    add_by_filehandles(single=True)
    add_by_bytes(single=False)
    add_by_bytes(single=True)


def test_add_torrent_file_fail(client, monkeypatch):
    with pytest.raises(TorrentFileNotFoundError):
        client.torrents_add(torrent_files="/tmp/asdfasdfasdfasdf")

    with pytest.raises(TorrentFilePermissionError):
        client.torrents_add(torrent_files="/etc/shadow")

    if platform.python_implementation() == "CPython":
        with pytest.raises(TorrentFileError):

            def fake_open(*arg, **kwargs):
                raise OSError(errno.ENODEV)

            with monkeypatch.context() as m:
                m.setitem(__builtins__, "open", fake_open)
                client.torrents_add(torrent_files="/etc/hosts")


@pytest.mark.parametrize("keep_root_folder", [True, False, None])
@pytest.mark.parametrize(
    "content_layout", [None, "Original", "Subfolder", "NoSubfolder"]
)
def test_add_options(client, api_version, keep_root_folder, content_layout, tmp_path):
    if v(api_version) >= v("2.3.0"):
        client.torrents_create_tags("option-tag")
    new_torrent = new_torrent_standalone(
        client=client,
        torrent_files=ROOT_FOLDER_TORRENT_FILE,
        torrent_hash=ROOT_FOLDER_TORRENT_HASH,
        save_path=mkpath(tmp_path, "test_download"),
        category="test_category",
        is_paused=True,
        upload_limit=1024,
        download_limit=2048,
        is_sequential_download=True,
        is_first_last_piece_priority=True,
        is_root_folder=keep_root_folder,
        rename="this is a new name for the torrent",
        use_auto_torrent_management=False,
        tags="option-tag",
        content_layout=content_layout,
        ratio_limit=2,
        seeding_time_limit=120,
    )

    with new_torrent as torrent:
        check(lambda: torrent.info.category, "test_category")
        check(
            lambda: torrent.info.state,
            ("pausedDL", "checkingResumeData"),
            reverse=True,
            any=True,
        )
        check(lambda: mkpath(torrent.info.save_path), mkpath(tmp_path, "test_download"))
        check(lambda: torrent.info.up_limit, 1024)
        check(lambda: torrent.info.dl_limit, 2048)
        check(lambda: torrent.info.seq_dl, True)
        if v(api_version) >= v("2.0.1"):
            check(lambda: torrent.info.f_l_piece_prio, True)
        if content_layout is None:
            check(
                lambda: torrent.files[0]["name"].startswith("root_folder"),
                keep_root_folder in {True, None},
            )
        check(lambda: torrent.info.name, "this is a new name for the torrent")
        check(lambda: torrent.info.auto_tmm, False)
        if v(api_version) >= v("2.6.2"):
            check(lambda: torrent.info.tags, "option-tag")

        if v(api_version) >= v("2.7"):
            # after web api v2.7...root dir is driven by content_layout
            if content_layout is None:
                should_root_dir_exists = keep_root_folder in {None, True}
            else:
                should_root_dir_exists = content_layout in {"Original", "Subfolder"}
        else:
            # before web api v2.7...it is driven by is_root_folder
            if content_layout is not None and keep_root_folder is None:
                should_root_dir_exists = content_layout in {"Original", "Subfolder"}
            else:
                should_root_dir_exists = keep_root_folder in {None, True}
        check(
            lambda: any(f["name"].startswith("root_folder") for f in torrent.files),
            should_root_dir_exists,
        )

        if v(api_version) >= v("2.8.1"):
            check(lambda: torrent.info.ratio_limit, 2)
            check(lambda: torrent.info.seeding_time_limit, 120)


@pytest.mark.skipif_before_api_version("2.8.4")
@pytest.mark.parametrize("use_download_path", [None, True, False])
def test_torrents_add_download_path(client, use_download_path, tmp_path):
    client.torrents_delete(torrent_hashes=ROOT_FOLDER_TORRENT_HASH, delete_files=True)
    save_path = mkpath(tmp_path, "down_path_save_path_test")
    download_path = mkpath(tmp_path, "down_path_test")
    new_torrent = new_torrent_standalone(
        client=client,
        torrent_hash=ROOT_FOLDER_TORRENT_HASH,
        torrent_files=ROOT_FOLDER_TORRENT_FILE,
        download_path=download_path,
        use_download_path=use_download_path,
        test_download_limit=1024,
        save_path=save_path,
    )

    with new_torrent as torrent:
        if use_download_path is False:
            check(
                lambda: mkpath(torrent.info.download_path), download_path, negate=True
            )
        else:
            check(lambda: mkpath(torrent.info.download_path), download_path)


@pytest.mark.skipif_before_api_version("2.9.3")
@pytest.mark.parametrize("count_func", ["torrents_count", "torrents.count"])
def test_count(client, count_func):
    assert client.func(count_func)() == 1


@pytest.mark.skipif_after_api_version("2.9.3")
@pytest.mark.parametrize("count_func", ["torrents_count", "torrents.count"])
def test_count_not_implemented(client, count_func):
    with pytest.raises(NotImplementedError):
        assert client.func(count_func)() == 1


@pytest.mark.parametrize(
    "properties_func", ["torrents_properties", "torrents.properties"]
)
def test_properties(client, orig_torrent, properties_func):
    props = client.func(properties_func)(torrent_hash=orig_torrent.hash)
    assert isinstance(props, TorrentPropertiesDictionary)


@pytest.mark.parametrize("trackers_func", ["torrents_trackers", "torrents.trackers"])
def test_trackers(client, orig_torrent, trackers_func):
    trackers = client.func(trackers_func)(torrent_hash=orig_torrent.hash)
    assert isinstance(trackers, TrackersList)


@pytest.mark.parametrize("trackers_func", ["torrents_trackers", "torrents.trackers"])
def test_trackers_slice(client, orig_torrent, trackers_func):
    trackers = client.func(trackers_func)(torrent_hash=orig_torrent.hash)
    assert isinstance(trackers[1:2], TrackersList)


@pytest.mark.parametrize("webseeds_func", ["torrents_webseeds", "torrents.webseeds"])
def test_webseeds(client, orig_torrent, webseeds_func):
    web_seeds = client.func(webseeds_func)(torrent_hash=orig_torrent.hash)
    assert isinstance(web_seeds, WebSeedsList)


@pytest.mark.parametrize("webseeds_func", ["torrents_webseeds", "torrents.webseeds"])
def test_webseeds_slice(client, orig_torrent, webseeds_func):
    web_seeds = client.func(webseeds_func)(torrent_hash=orig_torrent.hash)
    assert isinstance(web_seeds[1:2], WebSeedsList)


@pytest.mark.parametrize("files_func", ["torrents_files", "torrents.files"])
def test_files(client, orig_torrent, files_func):
    files = client.func(files_func)(torrent_hash=orig_torrent.hash)
    assert isinstance(files, TorrentFilesList)
    assert "availability" in files[0]
    assert all(file["id"] == file["index"] for file in files)


@pytest.mark.parametrize("files_func", ["torrents_files", "torrents.files"])
def test_files_slice(client, orig_torrent, files_func):
    files = client.func(files_func)(torrent_hash=orig_torrent.hash)
    assert isinstance(files[1:2], TorrentFilesList)


@pytest.mark.parametrize(
    "piece_state_func",
    [
        "torrents_piece_states",
        "torrents_pieceStates",
        "torrents.piece_states",
        "torrents.pieceStates",
    ],
)
def test_piece_states(client, orig_torrent, piece_state_func):
    piece_states = client.func(piece_state_func)(torrent_hash=orig_torrent.hash)
    assert isinstance(piece_states, TorrentPieceInfoList)


@pytest.mark.parametrize(
    "piece_state_func",
    [
        "torrents_piece_states",
        "torrents_pieceStates",
        "torrents.piece_states",
        "torrents.pieceStates",
    ],
)
def test_piece_states_slice(client, orig_torrent, piece_state_func):
    piece_states = client.func(piece_state_func)(torrent_hash=orig_torrent.hash)
    assert isinstance(piece_states[1:2], TorrentPieceInfoList)


@pytest.mark.parametrize(
    "piece_hashes_func",
    [
        "torrents_piece_hashes",
        "torrents_pieceHashes",
        "torrents.piece_hashes",
        "torrents.pieceHashes",
    ],
)
def test_piece_hashes(client, orig_torrent, piece_hashes_func):
    piece_hashes = client.func(piece_hashes_func)(torrent_hash=orig_torrent.hash)
    assert isinstance(piece_hashes, TorrentPieceInfoList)


@pytest.mark.parametrize(
    "piece_hashes_func",
    [
        "torrents_piece_hashes",
        "torrents_pieceHashes",
        "torrents.piece_hashes",
        "torrents.pieceHashes",
    ],
)
def test_piece_hashes_slice(client, orig_torrent, piece_hashes_func):
    piece_hashes = client.func(piece_hashes_func)(torrent_hash=orig_torrent.hash)
    assert isinstance(piece_hashes[1:2], TorrentPieceInfoList)


@pytest.mark.parametrize("trackers", ["127.0.0.1", ["127.0.0.2", "127.0.0.3"]])
@pytest.mark.parametrize(
    "add_trackers_func",
    [
        "torrents_add_trackers",
        "torrents_addTrackers",
        "torrents.add_trackers",
        "torrents.addTrackers",
    ],
)
def test_add_trackers(client, trackers, new_torrent, add_trackers_func):
    client.func(add_trackers_func)(torrent_hash=new_torrent.hash, urls=trackers)
    check(lambda: (t.url for t in new_torrent.trackers), trackers, reverse=True)


@pytest.mark.skipif_before_api_version("2.2.0")
@pytest.mark.parametrize(
    "edit_trackers_func",
    [
        "torrents_edit_tracker",
        "torrents_editTracker",
        "torrents.edit_tracker",
        "torrents.editTracker",
    ],
)
def test_edit_tracker(client, orig_torrent, edit_trackers_func):
    orig_torrent.add_trackers("127.1.0.1")
    client.func(edit_trackers_func)(
        torrent_hash=orig_torrent.hash,
        original_url="127.1.0.1",
        new_url="127.1.0.2",
    )
    check(lambda: (t.url for t in orig_torrent.trackers), "127.1.0.2", reverse=True)
    client.torrents_remove_trackers(torrent_hash=orig_torrent.hash, urls="127.1.0.2")


@pytest.mark.skipif_after_api_version("2.2.0")
@pytest.mark.parametrize(
    "edit_trackers_func",
    [
        "torrents_edit_tracker",
        "torrents_editTracker",
        "torrents.edit_tracker",
        "torrents.editTracker",
    ],
)
def test_edit_tracker_not_implemented(client, orig_torrent, edit_trackers_func):
    with pytest.raises(NotImplementedError):
        client.func(edit_trackers_func)()


@pytest.mark.skipif_before_api_version("2.2.0")
@pytest.mark.parametrize(
    "trackers",
    [
        ["127.2.0.1"],
        ["127.2.0.2", "127.2.0.3"],
    ],
)
@pytest.mark.parametrize(
    "remove_trackers_func",
    [
        "torrents_remove_trackers",
        "torrents_removeTrackers",
        "torrents.remove_trackers",
        "torrents.removeTrackers",
    ],
)
def test_remove_trackers(client, trackers, orig_torrent, remove_trackers_func):
    orig_torrent.add_trackers(trackers)
    client.func(remove_trackers_func)(torrent_hash=orig_torrent.hash, urls=trackers)
    check(
        lambda: (t.url for t in orig_torrent.trackers),
        trackers,
        reverse=True,
        negate=True,
    )


@pytest.mark.skipif_after_api_version("2.2.0")
@pytest.mark.parametrize(
    "remove_trackers_func",
    [
        "torrents_remove_trackers",
        "torrents_removeTrackers",
        "torrents.remove_trackers",
        "torrents.removeTrackers",
    ],
)
def test_remove_trackers_not_implemented(client, orig_torrent, remove_trackers_func):
    with pytest.raises(NotImplementedError):
        client.func(remove_trackers_func)()


@pytest.mark.parametrize(
    "file_prio_func",
    [
        "torrents_file_priority",
        "torrents_filePrio",
        "torrents.file_priority",
        "torrents.filePrio",
    ],
)
def test_file_priority(client, orig_torrent, file_prio_func):
    client.func(file_prio_func)(torrent_hash=orig_torrent.hash, file_ids=0, priority=6)
    check(lambda: orig_torrent.files[0].priority, 6)
    client.func(file_prio_func)(torrent_hash=orig_torrent.hash, file_ids=0, priority=7)
    check(lambda: orig_torrent.files[0].priority, 7)


@pytest.mark.parametrize("new_name", ["new name 2", "new_name_2"])
@pytest.mark.parametrize("rename_func", ["torrents_rename", "torrents.rename"])
def test_rename(client, new_torrent, new_name, rename_func):
    client.func(rename_func)(torrent_hash=new_torrent.hash, new_torrent_name=new_name)
    check(lambda: new_torrent.info.name.replace("+", " "), new_name)


@pytest.mark.skipif_before_api_version("2.4.0")
@pytest.mark.parametrize("new_name", ["new name file 2", "new_name_file_2"])
@pytest.mark.parametrize(
    "rename_file_func",
    [
        "torrents_rename_file",
        "torrents_renameFile",
        "torrents.rename_file",
        "torrents.renameFile",
    ],
)
def test_rename_file(
    client,
    new_torrent,
    new_name,
    rename_file_func,
):
    # pre-v4.3.3 rename_file signature
    client.func(rename_file_func)(
        torrent_hash=new_torrent.hash, file_id=0, new_file_name=new_name
    )
    check(lambda: new_torrent.files[0].name.replace("+", " "), new_name)
    # test invalid file ID is rejected
    with pytest.raises(Conflict409Error):
        client.func(rename_file_func)(
            torrent_hash=new_torrent.hash, file_id=10, new_file_name=new_name
        )
    # post-v4.3.3 rename_file signature
    new_new_name = new_name + "NEW"
    client.func(rename_file_func)(
        torrent_hash=new_torrent.hash,
        old_path=new_torrent.files[0].name,
        new_path=new_new_name,
    )
    check(lambda: new_torrent.files[0].name.replace("+", " "), new_new_name)
    # test invalid old_path is rejected
    with pytest.raises(Conflict409Error):
        client.func(rename_file_func)(
            torrent_hash=new_torrent.hash, old_path="asdf", new_path="xcvb"
        )


@pytest.mark.skipif_after_api_version("2.4.0")
@pytest.mark.parametrize(
    "rename_file_func",
    [
        "torrents_rename_file",
        "torrents_renameFile",
        "torrents.rename_file",
        "torrents.renameFile",
    ],
)
def test_rename_file_not_implemented(
    client,
    new_torrent,
    rename_file_func,
):
    with pytest.raises(NotImplementedError):
        client.func(rename_file_func)()


@pytest.mark.skipif_before_api_version("2.7")
@pytest.mark.parametrize("new_name", ["asdf zxcv", "asdf_zxcv"])
@pytest.mark.parametrize(
    "rename_folder_func",
    [
        "torrents_rename_folder",
        "torrents_renameFolder",
        "torrents.rename_folder",
        "torrents.renameFolder",
    ],
)
def test_rename_folder(client, app_version, new_torrent, new_name, rename_folder_func):
    if v(app_version) >= v("v4.3.3"):
        # move the file in to a new folder
        orig_file_path = new_torrent.files[0].name
        new_folder = "qwer"
        client.torrents_rename_file(
            torrent_hash=new_torrent.hash,
            old_path=orig_file_path,
            new_path=new_folder + "/" + orig_file_path,
        )

        # wait for the folder to be renamed
        check(
            lambda: [f.name.split("/")[0] for f in new_torrent.files],
            new_folder,
            reverse=True,
        )

        # test rename that new folder
        client.func(rename_folder_func)(
            torrent_hash=new_torrent.hash,
            old_path=new_folder,
            new_path=new_name,
        )
        check(
            lambda: new_torrent.files[0].name.replace("+", " "),
            new_name + "/" + orig_file_path,
        )
    elif v(app_version) >= v("v4.3.2"):
        with pytest.raises(NotImplementedError):
            client.func(rename_folder_func)()


@pytest.mark.skipif_after_api_version("2.7")
@pytest.mark.parametrize(
    "rename_folder_func",
    [
        "torrents_rename_folder",
        "torrents_renameFolder",
        "torrents.rename_folder",
        "torrents.renameFolder",
    ],
)
def test_rename_folder_not_implemented(client, rename_folder_func):
    with pytest.raises(NotImplementedError):
        client.func(rename_folder_func)()


@pytest.mark.skipif_before_api_version("2.8.14")
@pytest.mark.parametrize("export_func", ["torrents_export", "torrents.export"])
def test_export(client, orig_torrent, export_func):
    assert isinstance(client.func(export_func)(torrent_hash=orig_torrent.hash), bytes)


@pytest.mark.skipif_after_api_version("2.8.14")
@pytest.mark.parametrize("export_func", ["torrents_export", "torrents.export"])
def test_export_not_implemented(client, export_func):
    with pytest.raises(NotImplementedError):
        client.func(export_func)()


@pytest.mark.parametrize("info_func", ["torrents_info", "torrents.info"])
def test_torrents_info(client, info_func):
    assert isinstance(client.func(info_func)(), TorrentInfoList)
    if "." in info_func:
        assert isinstance(client.func(info_func).all(), TorrentInfoList)
        assert isinstance(client.func(info_func).downloading(), TorrentInfoList)
        assert isinstance(client.func(info_func).seeding(), TorrentInfoList)
        assert isinstance(client.func(info_func).completed(), TorrentInfoList)
        assert isinstance(client.func(info_func).paused(), TorrentInfoList)
        assert isinstance(client.func(info_func).active(), TorrentInfoList)
        assert isinstance(client.func(info_func).inactive(), TorrentInfoList)
        assert isinstance(client.func(info_func).resumed(), TorrentInfoList)
        assert isinstance(client.func(info_func).stalled(), TorrentInfoList)
        assert isinstance(client.func(info_func).stalled_uploading(), TorrentInfoList)
        assert isinstance(client.func(info_func).stalled_downloading(), TorrentInfoList)
        assert isinstance(client.func(info_func).checking(), TorrentInfoList)
        assert isinstance(client.func(info_func).moving(), TorrentInfoList)
        assert isinstance(client.func(info_func).errored(), TorrentInfoList)


def test_torrents_info_slice(client):
    assert isinstance(client.torrents_info()[1:2], TorrentInfoList)


@pytest.mark.skipif_before_api_version("2.8.3")
@pytest.mark.parametrize("info_func", ["torrents_info", "torrents.info"])
def test_torrents_info_tag(client, new_torrent, info_func):
    tag_name = "tag_filter_name"
    client.torrents_add_tags(tags=tag_name, torrent_hashes=new_torrent.hash)
    torrents = client.func(info_func)(torrent_hashes=new_torrent.hash, tag=tag_name)
    assert new_torrent.hash in {t.hash for t in torrents}


# test fails on 4.1.0 release
@pytest.mark.skipif_before_api_version("2.0.1")
@pytest.mark.parametrize(
    "pause_func, resume_func",
    [("torrents_pause", "torrents_resume"), ("torrents.pause", "torrents.resume")],
)
def test_pause_resume(client, new_torrent, pause_func, resume_func):
    client.func(pause_func)(torrent_hashes=new_torrent.hash)
    check(
        lambda: client.torrents_info(torrent_hashes=new_torrent.hash)[
            0
        ].state_enum.is_paused,
        True,
    )

    client.func(resume_func)(torrent_hashes=new_torrent.hash)
    check(
        lambda: client.torrents_info(torrent_hashes=new_torrent.hash)[
            0
        ].state_enum.is_paused,
        False,
    )


def test_action_for_all_torrents(client):
    client.torrents.resume.all()
    for torrent in client.torrents.info():
        check(
            lambda: client.torrents_info(torrent_hashes=torrent.hash)[0].state,
            {"pausedDL"},
            negate=True,
        )
    client.torrents.pause.all()
    for torrent in client.torrents.info():
        check(
            lambda: client.torrents_info(torrent_hashes=torrent.hash)[0].state,
            {"stalledDL", "pausedDL"},
            any=True,
        )


@pytest.mark.parametrize("recheck_func", ["torrents_recheck", "torrents.recheck"])
def test_recheck(client, orig_torrent, recheck_func):
    client.func(recheck_func)(torrent_hashes=orig_torrent.hash)


@pytest.mark.skipif_before_api_version("2.0.2")
@pytest.mark.parametrize(
    "reannounce_func", ["torrents_reannounce", "torrents.reannounce"]
)
def test_reannounce(client, orig_torrent, reannounce_func):
    client.func(reannounce_func)(torrent_hashes=orig_torrent.hash)


@pytest.mark.skipif_after_api_version("2.0.2")
@pytest.mark.parametrize(
    "reannounce_func", ["torrents_reannounce", "torrents.reannounce"]
)
def test_reannounce_not_implemented(client, reannounce_func):
    with pytest.raises(NotImplementedError):
        client.func(reannounce_func)()


# priority doesn't seem to work on v4.1.0
@pytest.mark.skipif_before_api_version("2.0.1")
@pytest.mark.parametrize(
    "inc_prio_func, dec_prio_func, top_prio_func, bottom_prio_func",
    [
        (
            "torrents_increase_priority",
            "torrents_decrease_priority",
            "torrents_top_priority",
            "torrents_bottom_priority",
        ),
        (
            "torrents_increasePrio",
            "torrents_decreasePrio",
            "torrents_topPrio",
            "torrents_bottomPrio",
        ),
    ],
)
def test_priority(
    client, new_torrent, inc_prio_func, dec_prio_func, top_prio_func, bottom_prio_func
):
    disable_queueing(client)

    with pytest.raises(Conflict409Error):
        client.func(inc_prio_func)(torrent_hashes=new_torrent.hash)
    with pytest.raises(Conflict409Error):
        client.func(dec_prio_func)(torrent_hashes=new_torrent.hash)
    with pytest.raises(Conflict409Error):
        client.func(top_prio_func)(torrent_hashes=new_torrent.hash)
    with pytest.raises(Conflict409Error):
        client.func(bottom_prio_func)(torrent_hashes=new_torrent.hash)

    enable_queueing(client)

    @retry()
    def test1(current_priority):
        client.func(inc_prio_func)(torrent_hashes=new_torrent.hash)
        check(lambda: new_torrent.info.priority < current_priority, True)

    @retry()
    def test2(current_priority):
        client.func(dec_prio_func)(torrent_hashes=new_torrent.hash)
        check(lambda: new_torrent.info.priority > current_priority, True)

    @retry()
    def test3(current_priority):
        client.func(top_prio_func)(torrent_hashes=new_torrent.hash)
        check(lambda: new_torrent.info.priority < current_priority, True)

    @retry()
    def test4(current_priority):
        client.func(bottom_prio_func)(torrent_hashes=new_torrent.hash)
        check(lambda: new_torrent.info.priority > current_priority, True)

    test1(current_priority=new_torrent.info.priority)
    test2(current_priority=new_torrent.info.priority)
    test3(current_priority=new_torrent.info.priority)
    test4(current_priority=new_torrent.info.priority)


@pytest.mark.parametrize(
    "set_down_limit_func, down_limit_func",
    [
        ("torrents_set_download_limit", "torrents_download_limit"),
        ("torrents_setDownloadLimit", "torrents_downloadLimit"),
        ("torrents.set_download_limit", "torrents.download_limit"),
        ("torrents.setDownloadLimit", "torrents.downloadLimit"),
    ],
)
def test_download_limit(client, orig_torrent, set_down_limit_func, down_limit_func):
    orig_download_limit = client.func(down_limit_func)(
        torrent_hashes=orig_torrent.hash
    )[orig_torrent.hash]

    client.func(set_down_limit_func)(torrent_hashes=orig_torrent.hash, limit=100)
    assert isinstance(
        client.func(down_limit_func)(torrent_hashes=orig_torrent.hash),
        TorrentLimitsDictionary,
    )
    check(
        lambda: client.func(down_limit_func)(torrent_hashes=orig_torrent.hash)[
            orig_torrent.hash
        ],
        100,
    )

    # reset download limit
    client.func(set_down_limit_func)(
        torrent_hashes=orig_torrent.hash, limit=orig_download_limit
    )
    check(
        lambda: client.func(down_limit_func)(torrent_hashes=orig_torrent.hash)[
            orig_torrent.hash
        ],
        orig_download_limit,
    )


@pytest.mark.parametrize(
    "set_up_limit_func, up_limit_func",
    [
        ("torrents_set_upload_limit", "torrents_upload_limit"),
        ("torrents_setUploadLimit", "torrents_uploadLimit"),
        ("torrents.set_upload_limit", "torrents.upload_limit"),
        ("torrents.setUploadLimit", "torrents.uploadLimit"),
    ],
)
def test_upload_limit(client, orig_torrent, set_up_limit_func, up_limit_func):
    orig_upload_limit = client.func(up_limit_func)(torrent_hashes=orig_torrent.hash)[
        orig_torrent.hash
    ]

    client.func(set_up_limit_func)(torrent_hashes=orig_torrent.hash, limit=100)
    assert isinstance(
        client.func(up_limit_func)(torrent_hashes=orig_torrent.hash),
        TorrentLimitsDictionary,
    )
    check(
        lambda: client.func(up_limit_func)(torrent_hashes=orig_torrent.hash)[
            orig_torrent.hash
        ],
        100,
    )

    # reset upload limit
    client.func(set_up_limit_func)(
        torrent_hashes=orig_torrent.hash, limit=orig_upload_limit
    )
    check(
        lambda: client.func(up_limit_func)(torrent_hashes=orig_torrent.hash)[
            orig_torrent.hash
        ],
        orig_upload_limit,
    )


@pytest.mark.skipif_before_api_version("2.0.1")
@pytest.mark.parametrize(
    "set_share_limits_func",
    [
        "torrents_set_share_limits",
        "torrents_setShareLimits",
        "torrents.set_share_limits",
        "torrents.setShareLimits",
    ],
)
def test_set_share_limits(client, orig_torrent, set_share_limits_func):
    client.func(set_share_limits_func)(
        ratio_limit=2,
        seeding_time_limit=5,
        inactive_seeding_time_limit=8,
        torrent_hashes=orig_torrent.hash,
    )
    check(lambda: orig_torrent.info.max_ratio, 2)
    check(lambda: orig_torrent.info.max_seeding_time, 5)
    if "max_inactive_seeding_time" in orig_torrent.info:
        check(lambda: orig_torrent.info.max_inactive_seeding_time, 8)

    client.func(set_share_limits_func)(
        ratio_limit=3,
        seeding_time_limit=6,
        inactive_seeding_time_limit=9,
        torrent_hashes=orig_torrent.hash,
    )
    check(lambda: orig_torrent.info.max_ratio, 3)
    check(lambda: orig_torrent.info.max_seeding_time, 6)
    if "max_inactive_seeding_time" in orig_torrent.info:
        check(lambda: orig_torrent.info.max_inactive_seeding_time, 9)


@pytest.mark.skipif_after_api_version("2.0.1")
@pytest.mark.parametrize(
    "set_share_limits_func",
    [
        "torrents_set_share_limits",
        "torrents_setShareLimits",
        "torrents.set_share_limits",
        "torrents.setShareLimits",
    ],
)
def test_set_share_limits_not_implemented(client, set_share_limits_func):
    with pytest.raises(NotImplementedError):
        client.func(set_share_limits_func)()


@pytest.mark.skipif_before_api_version("2.0.2")
@pytest.mark.parametrize(
    "set_loc_func",
    [
        "torrents_set_location",
        "torrents_setLocation",
        "torrents.set_location",
        "torrents.setLocation",
    ],
)
def test_set_location(client, app_version, new_torrent, set_loc_func, tmp_path):
    # stopped erroring when the write check was removed for API
    if v(app_version) < v("v4.5.2"):
        with pytest.raises(Forbidden403Error):
            client.func(set_loc_func)(location="/etc/", torrent_hashes=new_torrent.hash)

    sleep(0.5)
    loc = mkpath(tmp_path, "1")
    client.func(set_loc_func)(location=loc, torrent_hashes=new_torrent.hash)
    # qBittorrent may return trailing separators depending on version....
    check(lambda: mkpath(new_torrent.info.save_path), loc, any=True)


@pytest.mark.skipif_before_api_version("2.8.4")
@pytest.mark.parametrize(
    "set_save_path_func",
    [
        "torrents_set_save_path",
        "torrents_setSavePath",
        "torrents.set_save_path",
        "torrents.setSavePath",
    ],
)
def test_set_save_path(client, new_torrent, set_save_path_func, tmp_path):
    with pytest.raises(Forbidden403Error):
        client.func(set_save_path_func)(
            save_path="/etc/", torrent_hashes=new_torrent.hash
        )
    with pytest.raises(Conflict409Error):
        client.func(set_save_path_func)(
            save_path="/etc/asdf", torrent_hashes=new_torrent.hash
        )

    loc = mkpath(tmp_path, "savepath1")
    client.func(set_save_path_func)(save_path=loc, torrent_hashes=new_torrent.hash)
    # qBittorrent may return trailing separators depending on version....
    check(lambda: mkpath(new_torrent.info.save_path), loc, any=True)


@pytest.mark.skipif_after_api_version("2.8.4")
@pytest.mark.parametrize(
    "set_save_path_func",
    [
        "torrents_set_save_path",
        "torrents_setSavePath",
        "torrents.set_save_path",
        "torrents.setSavePath",
    ],
)
def test_set_save_path_not_implemented(client, set_save_path_func):
    with pytest.raises(NotImplementedError):
        client.func(set_save_path_func)()


@pytest.mark.skipif_before_api_version("2.8.4")
@pytest.mark.parametrize(
    "set_down_path_func",
    [
        "torrents_set_download_path",
        "torrents_setDownloadPath",
        "torrents.set_download_path",
        "torrents.setDownloadPath",
    ],
)
def test_set_download_path(client, new_torrent, set_down_path_func, tmp_path):
    with pytest.raises(Forbidden403Error):
        client.func(set_down_path_func)(
            download_path="/etc/", torrent_hashes=new_torrent.hash
        )
    with pytest.raises(Conflict409Error):
        client.func(set_down_path_func)(
            download_path="/etc/asdf", torrent_hashes=new_torrent.hash
        )

    loc = mkpath(tmp_path, "savepath1")
    client.func(set_down_path_func)(download_path=loc, torrent_hashes=new_torrent.hash)
    # qBittorrent may return trailing separators depending on version....
    check(lambda: mkpath(new_torrent.info.download_path), loc, any=True)


@pytest.mark.skipif_after_api_version("2.8.4")
@pytest.mark.parametrize(
    "set_down_path_func",
    [
        "torrents_set_download_path",
        "torrents_setDownloadPath",
        "torrents.set_download_path",
        "torrents.setDownloadPath",
    ],
)
def test_set_download_path_not_implemented(client, new_torrent, set_down_path_func):
    with pytest.raises(NotImplementedError):
        client.func(set_down_path_func)()


@pytest.mark.parametrize(
    "set_cat_func",
    [
        "torrents_set_category",
        "torrents_setCategory",
        "torrents.set_category",
        "torrents.setCategory",
    ],
)
@pytest.mark.parametrize("name", ["awesome cat", "awesome_cat"])
def test_set_category(client, orig_torrent, set_cat_func, name):
    with pytest.raises(Conflict409Error):
        client.func(set_cat_func)(
            category="/!@#$%^&*(", torrent_hashes=orig_torrent.hash
        )

    client.torrents_create_category(name=name)
    try:
        client.func(set_cat_func)(category=name, torrent_hashes=orig_torrent.hash)
        check(lambda: orig_torrent.info.category.replace("+", " "), name)
    finally:
        client.torrents_remove_categories(categories=name)


@pytest.mark.parametrize(
    "set_auto_mgmt_func",
    [
        "torrents_set_auto_management",
        "torrents_setAutoManagement",
        "torrents.set_auto_management",
        "torrents.setAutoManagement",
    ],
)
def test_torrents_set_auto_management(client, orig_torrent, set_auto_mgmt_func):
    current_setting = orig_torrent.info.auto_tmm
    client.func(set_auto_mgmt_func)(
        enable=(not current_setting), torrent_hashes=orig_torrent.hash
    )
    check(lambda: orig_torrent.info.auto_tmm, (not current_setting))
    client.func(set_auto_mgmt_func)(
        enable=False, torrent_hashes=orig_torrent.hash
    )  # leave on False


@pytest.mark.parametrize(
    "toggle_seq_down_func",
    [
        "torrents_toggle_sequential_download",
        "torrents_toggleSequentialDownload",
        "torrents.toggle_sequential_download",
        "torrents.toggleSequentialDownload",
    ],
)
def test_toggle_sequential_download(client, orig_torrent, toggle_seq_down_func):
    current_setting = orig_torrent.info.seq_dl
    client.func(toggle_seq_down_func)(torrent_hashes=orig_torrent.hash)
    check(lambda: orig_torrent.info.seq_dl, not current_setting)


@pytest.mark.skipif_before_api_version("2.0.1")
@pytest.mark.parametrize(
    "toggle_piece_prio_func",
    [
        "torrents_toggle_first_last_piece_priority",
        "torrents_toggleFirstLastPiecePrio",
        "torrents.toggle_first_last_piece_priority",
        "torrents.toggleFirstLastPiecePrio",
    ],
)
def test_toggle_first_last_piece_priority(client, new_torrent, toggle_piece_prio_func):
    current_setting = new_torrent.info.f_l_piece_prio
    client.func(toggle_piece_prio_func)(torrent_hashes=new_torrent.hash)
    check(lambda: new_torrent.info.f_l_piece_prio, not current_setting)


@pytest.mark.parametrize(
    "set_force_start_func",
    [
        "torrents_set_force_start",
        "torrents_setForceStart",
        "torrents.set_force_start",
        "torrents.setForceStart",
    ],
)
def test_set_force_start(client, orig_torrent, set_force_start_func):
    current_setting = orig_torrent.info.force_start
    client.func(set_force_start_func)(
        enable=(not current_setting), torrent_hashes=orig_torrent.hash
    )
    check(lambda: orig_torrent.info.force_start, not current_setting)


@pytest.mark.parametrize(
    "set_super_seeding_func",
    [
        "torrents_set_super_seeding",
        "torrents_setSuperSeeding",
        "torrents.set_super_seeding",
        "torrents.setSuperSeeding",
    ],
)
def test_set_super_seeding(client, orig_torrent, set_super_seeding_func):
    client.func(set_super_seeding_func)(enable=False, torrent_hashes=orig_torrent.hash)
    check(lambda: orig_torrent.info.force_start, False)


@pytest.mark.skipif_before_api_version("2.3.0")
@pytest.mark.parametrize(
    "add_peers_func",
    [
        "torrents_add_peers",
        "torrents_addPeers",
        "torrents.add_peers",
        "torrents.addPeers",
    ],
)
@pytest.mark.parametrize(
    "peers", ["127.0.0.1:5000", ["127.0.0.1:5000", "127.0.0.2:5000"], "127.0.0.1"]
)
def test_torrents_add_peers(client, orig_torrent, add_peers_func, peers):
    if all(":" not in p for p in peers):
        with pytest.raises(InvalidRequest400Error):
            client.func(add_peers_func)(peers=peers, torrent_hashes=orig_torrent.hash)
    else:
        p = client.func(add_peers_func)(peers=peers, torrent_hashes=orig_torrent.hash)
        assert isinstance(p, TorrentsAddPeersDictionary)


@pytest.mark.skipif_after_api_version("2.3.0")
@pytest.mark.parametrize(
    "add_peers_func",
    [
        "torrents_add_peers",
        "torrents_addPeers",
        "torrents.add_peers",
        "torrents.addPeers",
    ],
)
def test_torrents_add_peers_not_implemented(client, add_peers_func):
    with pytest.raises(NotImplementedError):
        client.func(add_peers_func)()


def _categories_save_path_key(api_version):
    """With qBittorrent 4.4.0 (Web API 2.8.4), the key in the category definition
    returned changed from savePath to save_path...."""
    if v(api_version) == v("2.8.4"):
        return "save_path"
    return "savePath"


@pytest.mark.skipif_before_api_version("2.1.1")
def test_categories1(client):
    assert isinstance(client.torrents_categories(), TorrentCategoriesDictionary)


@pytest.mark.skipif_after_api_version("2.1.1")
def test_categories1_not_implemented(client, api_version):
    with pytest.raises(NotImplementedError):
        client.torrents_categories()


@pytest.mark.skipif_before_api_version("2.1.1")
def test_categories2(client, api_version, tmp_path):
    save_path_key = _categories_save_path_key(api_version)
    name = "new_category"
    client.torrent_categories.categories = {"name": name, save_path_key: tmp_path}
    assert name in client.torrent_categories.categories
    client.torrent_categories.categories = {
        "name": name,
        save_path_key: mkpath(tmp_path, "new"),
    }
    assert mkpath(client.torrent_categories.categories[name][save_path_key]) == mkpath(
        tmp_path, "new"
    )
    client.torrents_remove_categories(categories=name)


@pytest.mark.skipif_after_api_version("2.1.1")
def test_categories2_not_implemented(client):
    with pytest.raises(NotImplementedError):
        _ = client.torrent_categories.categories


@pytest.mark.parametrize(
    "create_cat_func",
    [
        "torrents_create_category",
        "torrents_createCategory",
        "torrent_categories.create_category",
        "torrent_categories.createCategory",
    ],
)
@pytest.mark.parametrize("filepath", [None, "", "/tmp/"])
@pytest.mark.parametrize("name", ["name", "name 1"])
@pytest.mark.parametrize("enable_download_path", [None, True, False])
def test_create_categories(
    client,
    api_version,
    orig_torrent,
    create_cat_func,
    filepath,
    name,
    enable_download_path,
):
    save_path = download_path = filepath
    if filepath:
        save_path += "save"
        download_path += "download"

    try:
        client.func(create_cat_func)(
            name=name,
            save_path=save_path,
            download_path=download_path,
            enable_download_path=enable_download_path,
        )
        client.torrents_set_category(torrent_hashes=orig_torrent.hash, category=name)
        check(lambda: orig_torrent.info.category.replace("+", " "), name)
        if v(api_version) >= v("2.2"):
            check(
                lambda: [n.replace("+", " ") for n in client.torrents_categories()],
                name,
                reverse=True,
            )
            save_path_key = _categories_save_path_key(api_version)
            check(
                lambda: [
                    mkpath(cat[save_path_key])
                    for cat in client.torrents_categories().values()
                ],
                mkpath(save_path) or "",
                reverse=True,
            )
        if v(api_version) >= v("2.8.4") and enable_download_path is not False:
            check(
                lambda: [
                    mkpath(cat.get("download_path", ""))
                    for cat in client.torrents_categories().values()
                ],
                mkpath(download_path) or "",
                reverse=True,
            )
    finally:
        client.torrents_remove_categories(categories=name)


@pytest.mark.skipif_before_api_version("2.1.0")
@pytest.mark.parametrize(
    "edit_cat_func",
    [
        "torrents_edit_category",
        "torrents_editCategory",
        "torrent_categories.edit_category",
        "torrent_categories.editCategory",
    ],
)
@pytest.mark.parametrize("filepath", ["", "/tmp/"])
@pytest.mark.parametrize("name", ["editcategory"])
@pytest.mark.parametrize("enable_download_path", [None, True, False])
def test_edit_category(
    client, api_version, edit_cat_func, filepath, name, enable_download_path
):
    try:
        client.torrents_create_category(
            name=name, save_path="/tmp/savetmp", download_path="/tmp/savetmp"
        )
        save_path = mkpath(filepath + "save/")
        download_path = mkpath(filepath + "down/")
        client.func(edit_cat_func)(
            name=name,
            save_path=save_path,
            download_path=download_path,
            enable_download_path=enable_download_path,
        )
        check(
            lambda: [n.replace("+", " ") for n in client.torrents_categories()],
            name,
            reverse=True,
        )
        save_path_key = _categories_save_path_key(api_version)
        check(
            lambda: (
                mkpath(cat[save_path_key])
                for cat in client.torrents_categories().values()
            ),
            mkpath(save_path) or "",
            reverse=True,
        )
        if v(api_version) >= v("2.8.4") and enable_download_path is not False:
            check(
                lambda: [
                    mkpath(cat.get("download_path", ""))
                    for cat in client.torrents_categories().values()
                ],
                mkpath(download_path) or "",
                reverse=True,
            )
    finally:
        client.torrents_remove_categories(categories=name)


@pytest.mark.skipif_after_api_version("2.1.0")
@pytest.mark.parametrize(
    "edit_cat_func",
    (
        "torrents_edit_category",
        "torrents_editCategory",
        "torrent_categories.edit_category",
        "torrent_categories.editCategory",
    ),
)
def test_edit_category_not_implemented(client, edit_cat_func):
    with pytest.raises(NotImplementedError):
        client.func(edit_cat_func)()


@pytest.mark.parametrize(
    "remove_cat_func",
    [
        "torrents_remove_categories",
        "torrents_removeCategories",
        "torrent_categories.remove_categories",
        "torrent_categories.removeCategories",
    ],
)
@pytest.mark.parametrize("categories", [["category1"], ["category1", "category 2"]])
def test_remove_category(
    client, api_version, orig_torrent, remove_cat_func, categories
):
    for name in categories:
        client.torrents_create_category(name=name)
    orig_torrent.set_category(category=categories[0])
    client.func(remove_cat_func)(categories=categories)
    if v(api_version) >= v("2.2"):
        check(
            lambda: [n.replace("+", " ") for n in client.torrents_categories()],
            categories,
            reverse=True,
            negate=True,
        )
    check(lambda: orig_torrent.info.category, categories[0], negate=True)


@pytest.mark.skipif_before_api_version("2.3.0")
@pytest.mark.parametrize(
    "tags_func",
    [
        "torrents_tags",
        "torrent_tags.tags",
    ],
)
def test_tags(client, tags_func):
    try:
        assert isinstance(client.func(tags_func)(), TagList)
    except TypeError:
        assert isinstance(client.func(tags_func), TagList)


@pytest.mark.skipif_before_api_version("2.3.0")
def test_tags_slice(client):
    assert isinstance(client.torrents_tags()[1:2], TagList)


@pytest.mark.skipif_after_api_version("2.3.0")
@pytest.mark.parametrize(
    "tags_func",
    [
        "torrents_tags",
        "torrent_tags.tags",
    ],
)
def test_tags_not_implemented(client, tags_func):
    with pytest.raises(NotImplementedError):
        client.func(tags_func)()


@pytest.mark.skipif_before_api_version("2.3.0")
def test_add_tag_though_property(client):
    name = "newtag"
    client.torrent_tags.tags = name
    assert name in client.torrent_tags.tags
    client.torrent_tags.delete_tags(name)
    assert name not in client.torrent_tags.tags


@pytest.mark.skipif_after_api_version("2.3.0")
def test_add_tag_though_property_not_implemented(client):
    with pytest.raises(NotImplementedError):
        client.torrent_tags.tags = "asdf"


@pytest.mark.skipif_before_api_version("2.3.0")
@pytest.mark.parametrize(
    "add_tags_func",
    [
        "torrents_add_tags",
        "torrents_addTags",
        "torrent_tags.add_tags",
        "torrent_tags.addTags",
    ],
)
@pytest.mark.parametrize("tags", [["tag1"], ["tag1", "tag 2"]])
def test_add_tags(client, orig_torrent, add_tags_func, tags):
    try:
        client.func(add_tags_func)(tags=tags, torrent_hashes=orig_torrent.hash)
        check(lambda: orig_torrent.info.tags, tags, reverse=True)
    finally:
        client.torrents_delete_tags(tags=tags)


@pytest.mark.skipif_after_api_version("2.3.0")
@pytest.mark.parametrize(
    "add_tags_func",
    [
        "torrents_add_tags",
        "torrents_addTags",
        "torrent_tags.add_tags",
        "torrent_tags.addTags",
    ],
)
def test_add_tags_not_implemented(client, add_tags_func):
    with pytest.raises(NotImplementedError):
        client.func(add_tags_func)()


@pytest.mark.skipif_before_api_version("2.3.0")
@pytest.mark.parametrize(
    "remove_tags_func",
    [
        "torrents_remove_tags",
        "torrents_removeTags",
        "torrent_tags.remove_tags",
        "torrent_tags.removeTags",
    ],
)
@pytest.mark.parametrize("tags", [["tag1"], ["tag1", "tag 2"]])
def test_remove_tags(client, orig_torrent, remove_tags_func, tags):
    try:
        orig_torrent.add_tags(tags=tags)
        client.func(remove_tags_func)(tags=tags, torrent_hashes=orig_torrent.hash)
        check(lambda: orig_torrent.info.tags, tags, reverse=True, negate=True)
    finally:
        client.torrents_delete_tags(tags=tags)


@pytest.mark.skipif_after_api_version("2.3.0")
@pytest.mark.parametrize(
    "remove_tags_func",
    [
        "torrents_remove_tags",
        "torrents_removeTags",
        "torrent_tags.remove_tags",
        "torrent_tags.removeTags",
    ],
)
def test_remove_tags_not_implemented(client, remove_tags_func):
    with pytest.raises(NotImplementedError):
        client.func(remove_tags_func)()


@pytest.mark.skipif_before_api_version("2.3.0")
@pytest.mark.parametrize(
    "create_tags_func",
    [
        "torrents_create_tags",
        "torrents_createTags",
        "torrent_tags.create_tags",
        "torrent_tags.createTags",
    ],
)
@pytest.mark.parametrize("tags", [["tag1"], ["tag1", "tag 2"]])
def test_create_tags(client, create_tags_func, tags):
    try:
        client.func(create_tags_func)(tags=tags)
        check(lambda: client.torrents_tags(), tags, reverse=True)
    finally:
        client.torrents_delete_tags(tags=tags)


@pytest.mark.skipif_after_api_version("2.3.0")
@pytest.mark.parametrize(
    "create_tags_func",
    [
        "torrents_create_tags",
        "torrents_createTags",
        "torrent_tags.create_tags",
        "torrent_tags.createTags",
    ],
)
def test_create_tags_not_implemented(client, create_tags_func):
    with pytest.raises(NotImplementedError):
        client.func(create_tags_func)()


@pytest.mark.skipif_before_api_version("2.3.0")
@pytest.mark.parametrize(
    "delete_tags_func",
    [
        "torrents_delete_tags",
        "torrents_deleteTags",
        "torrent_tags.delete_tags",
        "torrent_tags.deleteTags",
    ],
)
@pytest.mark.parametrize("tags", [["tag1"], ["tag1", "tag 2"]])
def test_delete_tags(client, delete_tags_func, tags):
    client.torrents_create_tags(tags=tags)
    client.func(delete_tags_func)(tags=tags)
    check(lambda: client.torrents_tags(), tags, reverse=True, negate=True)


@pytest.mark.skipif_after_api_version("2.3.0")
@pytest.mark.parametrize(
    "delete_tags_func",
    [
        "torrents_delete_tags",
        "torrents_deleteTags",
        "torrent_tags.delete_tags",
        "torrent_tags.deleteTags",
    ],
)
def test_delete_tags_not_implemented(client, delete_tags_func):
    with pytest.raises(NotImplementedError):
        client.func(delete_tags_func)()



from __future__ import annotations

import enum
import os
from collections.abc import Mapping
from functools import wraps
from typing import Any, Literal, cast

from qbittorrentapi.app import AppAPIMixIn
from qbittorrentapi.definitions import (
    APIKwargsT,
    APINames,
    ClientCache,
    Dictionary,
    JsonValueT,
    List,
    ListEntry,
    ListInputT,
)


class TaskStatus(enum.Enum):
    """Enumeration of possible task statuses."""

    FAILED = "Failed"
    QUEUED = "Queued"
    RUNNING = "Running"
    FINISHED = "Finished"


class TorrentCreatorTaskStatus(ListEntry):
    """
    Item in :class:`TorrentCreatorTaskStatusList`

    Definition: not documented...yet
    """


class TorrentCreatorTaskStatusList(List[TorrentCreatorTaskStatus]):
    """Response for :meth:`~TorrentCreatorAPIMixIn.torrentcreator_status`"""

    def __init__(
        self, list_entries: ListInputT, client: TorrentCreatorAPIMixIn | None = None
    ):
        super().__init__(
            list_entries, entry_class=TorrentCreatorTaskStatus, client=client
        )


class TorrentCreatorAPIMixIn(AppAPIMixIn):
    """
    Implementation of all ``TorrentCreator`` API methods.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host="localhost:8080", username="admin", password="adminadmin")
        >>> task = client.torrentcreator_add_task(source_path="/path/to/data")
        >>> if TaskStatus(task.status().status) == TaskStatus.FINISHED:
        >>>     torrent_data = task.torrent_file()
        >>> task.delete()
        >>> # or
        >>> client.torrentcreator_delete_task(task_id=task.taskID)
    """  # noqa: E501

    @property
    def torrentcreator(self) -> TorrentCreator:
        """
        Allows for transparent interaction with TorrentCreator endpoints.

        See TorrentCreator class for usage.
        """
        if self._torrentcreator is None:
            self._torrentcreator = TorrentCreator(client=self)
        return self._torrentcreator

    def torrentcreator_add_task(
        self,
        source_path: str | os.PathLike[Any] | None = None,
        torrent_file_path: str | os.PathLike[Any] | None = None,
        format: Literal["v1", "v2", "hybrid"] | None = None,
        start_seeding: bool | None = None,
        is_private: bool | None = None,
        optimize_alignment: bool | None = None,
        padded_file_size_limit: int | None = None,
        piece_size: int | None = None,
        comment: str | None = None,
        trackers: str | list[str] | None = None,
        url_seeds: str | list[str] | None = None,
        **kwargs: APIKwargsT,
    ) -> TorrentCreatorTaskDictionary:
        """
        Add a task to create a new torrent.

        This method was introduced with qBittorrent v5.0.0 (Web API v2.10.4).

        :raises Conflict409Error: too many existing torrent creator tasks
        :param source_path: source file path for torrent content
        :param torrent_file_path: file path to save torrent
        :param format: BitTorrent V1 or V2; defaults to "hybrid" format if None
        :param start_seeding: should qBittorrent start seeding this torrent?
        :param is_private: is the torrent private or not?
        :param optimize_alignment: should optimized alignment be enforced for new
            torrent?
        :param padded_file_size_limit: size limit for padding files
        :param piece_size: size of the pieces
        :param comment: comment
        :param trackers: list of trackers to add
        :param url_seeds: list of URLs seeds to add
        """
        data = {
            "sourcePath": source_path,
            "torrentFilePath": os.fsdecode(torrent_file_path)
            if torrent_file_path
            else None,
            "format": format,
            "private": None if is_private is None else bool(is_private),
            "optimizeAlignment": None
            if optimize_alignment is None
            else bool(optimize_alignment),
            "startSeeding": None if start_seeding is None else bool(start_seeding),
            "paddedFileSizeLimit": padded_file_size_limit,
            "pieceSize": piece_size,
            "comment": comment,
            "trackers": self._list2string(trackers),
            "urlSeeds": self._list2string(url_seeds),
        }

        return self._post_cast(
            _name=APINames.TorrentCreator,
            _method="addTask",
            data=data,
            response_class=TorrentCreatorTaskDictionary,
            version_introduced="2.10.4",
            **kwargs,
        )

    torrentcreator_addTask = torrentcreator_add_task

    def torrentcreator_status(
        self,
        task_id: str | None = None,
        **kwargs: APIKwargsT,
    ) -> TorrentCreatorTaskStatusList:
        """
        Status for a torrent creation task.

        This method was introduced with qBittorrent v5.0.0 (Web API v2.10.4).

        :raises NotFound404Error: task not found
        :param task_id: ID of torrent creation task
        """
        data = {"taskID": task_id}

        return self._post_cast(
            _name=APINames.TorrentCreator,
            _method="status",
            data=data,
            response_class=TorrentCreatorTaskStatusList,
            version_introduced="2.10.4",
            **kwargs,
        )

    def torrentcreator_torrent_file(
        self,
        task_id: str | None = None,
        **kwargs: APIKwargsT,
    ) -> bytes:
        """
        Retrieve torrent file for created torrent.

        This method was introduced with qBittorrent v5.0.0 (Web API v2.10.4).

        :raises NotFound404Error: task not found
        :raises Conflict409Error: torrent creation is not finished or failed
        :param task_id: ID of torrent creation task
        """
        data = {"taskID": task_id}

        return self._post_cast(
            _name=APINames.TorrentCreator,
            _method="torrentFile",
            data=data,
            response_class=bytes,
            version_introduced="2.10.4",
            **kwargs,
        )

    torrentcreator_torrentFile = torrentcreator_torrent_file

    def torrentcreator_delete_task(
        self,
        task_id: str | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """
        Delete a torrent creation task.

        This method was introduced with qBittorrent v5.0.0 (Web API v2.10.4).

        :raises NotFound404Error: task not found
        :param task_id: ID of torrent creation task
        """
        data = {"taskID": task_id}

        self._post(
            _name=APINames.TorrentCreator,
            _method="deleteTask",
            data=data,
            version_introduced="2.10.4",
            **kwargs,
        )

    torrentcreator_deleteTask = torrentcreator_delete_task


class TorrentCreatorTaskDictionary(
    ClientCache[TorrentCreatorAPIMixIn], Dictionary[JsonValueT]
):
    """Response for :meth:`~TorrentCreatorAPIMixIn.torrentcreator_add_task`"""

    def __init__(self, data: Mapping[str, JsonValueT], client: TorrentCreatorAPIMixIn):
        self.task_id: str | None = cast(str, data.get("taskID", None))
        super().__init__(data=data, client=client)

    @wraps(TorrentCreatorAPIMixIn.torrentcreator_status)
    def status(self, **kwargs: APIKwargsT) -> TorrentCreatorTaskStatus:
        return self._client.torrentcreator_status(task_id=self.task_id, **kwargs)[0]

    @wraps(TorrentCreatorAPIMixIn.torrentcreator_torrent_file)
    def torrent_file(self, **kwargs: APIKwargsT) -> bytes:
        return self._client.torrentcreator_torrent_file(task_id=self.task_id, **kwargs)

    torrentFile = torrent_file

    @wraps(TorrentCreatorAPIMixIn.torrentcreator_delete_task)
    def delete(self, **kwargs: APIKwargsT) -> None:
        return self._client.torrentcreator_delete_task(task_id=self.task_id, **kwargs)


class TorrentCreator(ClientCache[TorrentCreatorAPIMixIn]):
    """
    Allows interaction with ``TorrentCreator`` API endpoints.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host="localhost:8080", username="admin", password="adminadmin")
        >>> # this is all the same attributes that are available as named in the
        >>> #  endpoints or the more pythonic names in Client (with or without 'torrentcreator_' prepended)
        >>> task = client.torrentcreator.add_task(source_path="/path/to/data")
        >>> if TaskStatus(task.status().status) == TaskStatus.FINISHED:
        >>>     torrent_data = task.torrent_file()
        >>> task.delete()
        >>> # or
        >>> client.torrentcreator.delete_task(task_id=task.taskID)
    """  # noqa: E501

    @wraps(TorrentCreatorAPIMixIn.torrentcreator_add_task)
    def add_task(
        self,
        source_path: str | os.PathLike[Any] | None = None,
        torrent_file_path: str | os.PathLike[Any] | None = None,
        format: Literal["v1", "v2", "hybrid"] | None = None,
        start_seeding: bool | None = None,
        is_private: bool | None = None,
        optimize_alignment: bool | None = None,
        padded_file_size_limit: int | None = None,
        piece_size: int | None = None,
        comment: str | None = None,
        trackers: str | list[str] | None = None,
        url_seeds: str | list[str] | None = None,
        **kwargs: APIKwargsT,
    ) -> TorrentCreatorTaskDictionary:
        return self._client.torrentcreator_add_task(
            source_path=source_path,
            torrent_file_path=torrent_file_path,
            format=format,
            start_seeding=start_seeding,
            is_private=is_private,
            optimize_alignment=optimize_alignment,
            padded_file_size_limit=padded_file_size_limit,
            piece_size=piece_size,
            comment=comment,
            trackers=trackers,
            url_seeds=url_seeds,
            **kwargs,
        )

    addTask = add_task

    @wraps(TorrentCreatorAPIMixIn.torrentcreator_status)
    def status(
        self,
        task_id: str | None = None,
        **kwargs: APIKwargsT,
    ) -> TorrentCreatorTaskStatusList:
        return self._client.torrentcreator_status(task_id=task_id, **kwargs)

    @wraps(TorrentCreatorAPIMixIn.torrentcreator_torrent_file)
    def torrent_file(
        self,
        task_id: str | None = None,
        **kwargs: APIKwargsT,
    ) -> bytes:
        return self._client.torrentcreator_torrent_file(task_id=task_id, **kwargs)

    torrentFile = torrent_file

    @wraps(TorrentCreatorAPIMixIn.torrentcreator_delete_task)
    def delete_task(
        self,
        task_id: str | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        self._client.torrentcreator_delete_task(task_id=task_id, **kwargs)

    deleteTask = delete_task



from __future__ import annotations

from functools import wraps

from qbittorrentapi.app import AppAPIMixIn
from qbittorrentapi.definitions import (
    APIKwargsT,
    APINames,
    ClientCache,
    List,
    ListEntry,
    ListInputT,
)


class LogPeer(ListEntry):
    """Item in :class:`LogPeersList`"""


class LogPeersList(List[LogPeer]):
    """Response for :meth:`~LogAPIMixIn.log_peers`"""

    def __init__(self, list_entries: ListInputT, client: LogAPIMixIn | None = None):
        super().__init__(list_entries, entry_class=LogPeer)


class LogEntry(ListEntry):
    """Item in :class:`LogMainList`"""


class LogMainList(List[LogEntry]):
    """Response to :meth:`~LogAPIMixIn.log_main`"""

    def __init__(self, list_entries: ListInputT, client: LogAPIMixIn | None = None):
        super().__init__(list_entries, entry_class=LogEntry)


class LogAPIMixIn(AppAPIMixIn):
    """
    Implementation of all ``Log`` API methods.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host="localhost:8080", username="admin", password="adminadmin")
        >>> client.log_main(info=False)
        >>> client.log_peers()
    """  # noqa: E501

    @property
    def log(self) -> Log:
        """
        Allows for transparent interaction with Log endpoints.

        See Log class for usage.
        """
        if self._log is None:
            self._log = Log(client=self)
        return self._log

    def log_main(
        self,
        normal: bool | None = None,
        info: bool | None = None,
        warning: bool | None = None,
        critical: bool | None = None,
        last_known_id: str | int | None = None,
        **kwargs: APIKwargsT,
    ) -> LogMainList:
        """
        Retrieve the qBittorrent log entries. Iterate over returned object.

        :param normal: False to exclude ``normal`` entries
        :param info: False to exclude ``info`` entries
        :param warning: False to exclude ``warning`` entries
        :param critical: False to exclude ``critical`` entries
        :param last_known_id: only entries with an ID greater than this value will be
            returned
        """
        params = {
            "normal": None if normal is None else bool(normal),
            "info": None if info is None else bool(info),
            "warning": None if warning is None else bool(warning),
            "critical": None if critical is None else bool(critical),
            "last_known_id": last_known_id,
        }
        return self._get_cast(
            _name=APINames.Log,
            _method="main",
            params=params,
            response_class=LogMainList,
            **kwargs,
        )

    def log_peers(
        self,
        last_known_id: str | int | None = None,
        **kwargs: APIKwargsT,
    ) -> LogPeersList:
        """
        Retrieve qBittorrent peer log.

        :param last_known_id: only entries with an ID greater than this value will be
            returned
        """
        params = {"last_known_id": last_known_id}
        return self._get_cast(
            _name=APINames.Log,
            _method="peers",
            params=params,
            response_class=LogPeersList,
            **kwargs,
        )


class Log(ClientCache[LogAPIMixIn]):
    """
    Allows interaction with ``Log`` API endpoints.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host="localhost:8080", username="admin", password="adminadmin")
        >>> # this is all the same attributes that are available as named in the
        >>> #  endpoints or the more pythonic names in Client (with or without 'log_' prepended)
        >>> log_list = client.log.main()
        >>> peers_list = client.log.peers(last_known_id="...")
        >>> # can also filter log down with additional attributes
        >>> log_info = client.log.main.info(last_known_id=1)
        >>> log_warning = client.log.main.warning(last_known_id=1)
    """  # noqa: E501

    def __init__(self, client: LogAPIMixIn):
        super().__init__(client=client)
        self.main = Log._Main(client=client)

    @wraps(LogAPIMixIn.log_peers)
    def peers(
        self,
        last_known_id: str | int | None = None,
        **kwargs: APIKwargsT,
    ) -> LogPeersList:
        return self._client.log_peers(last_known_id=last_known_id, **kwargs)

    class _Main(ClientCache["LogAPIMixIn"]):
        def _api_call(
            self,
            normal: bool | None = None,
            info: bool | None = None,
            warning: bool | None = None,
            critical: bool | None = None,
            last_known_id: str | int | None = None,
            **kwargs: APIKwargsT,
        ) -> LogMainList:
            return self._client.log_main(
                normal=normal,
                info=info,
                warning=warning,
                critical=critical,
                last_known_id=last_known_id,
                **kwargs,
            )

        def __call__(
            self,
            normal: bool | None = True,
            info: bool | None = True,
            warning: bool | None = True,
            critical: bool | None = True,
            last_known_id: str | int | None = None,
            **kwargs: APIKwargsT,
        ) -> LogMainList:
            return self._api_call(
                normal=normal,
                info=info,
                warning=warning,
                critical=critical,
                last_known_id=last_known_id,
                **kwargs,
            )

        def info(
            self,
            last_known_id: str | int | None = None,
            **kwargs: APIKwargsT,
        ) -> LogMainList:
            return self._api_call(last_known_id=last_known_id, **kwargs)

        def normal(
            self,
            last_known_id: str | int | None = None,
            **kwargs: APIKwargsT,
        ) -> LogMainList:
            return self._api_call(info=False, last_known_id=last_known_id, **kwargs)

        def warning(
            self,
            last_known_id: str | int | None = None,
            **kwargs: APIKwargsT,
        ) -> LogMainList:
            return self._api_call(
                info=False,
                normal=False,
                last_known_id=last_known_id,
                **kwargs,
            )

        def critical(
            self,
            last_known_id: str | int | None = None,
            **kwargs: APIKwargsT,
        ) -> LogMainList:
            return self._api_call(
                info=False,
                normal=False,
                warning=False,
                last_known_id=last_known_id,
                **kwargs,
            )



# Copyright (c) 2013 Brendan Curran-Johnson
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Functions and Classes from attrdict.

*rmartin Feb 2021
AttrDict finally broke with Python 3.10 since abstract base classes can no
longer be imported from collections but should use collections.abc instead.
Since AttrDict is abandoned, I've consolidated the code here for future use.
AttrMap and AttrDefault were removed altogether.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from re import compile as re_compile
from typing import Any, Dict, Mapping, MutableMapping, Sequence, TypeVar

K = TypeVar("K")
V = TypeVar("V")
T = TypeVar("T")

VALID_KEY_RE = re_compile(r"^[A-Za-z][A-Za-z0-9_]*$")


def merge(left: Mapping[K, V], right: Mapping[K, V]) -> dict[K, V]:
    """
    Merge two mappings objects together, combining overlapping Mappings, and
    favoring right-values.

    left: The left Mapping object.
    right: The right (favored) Mapping object.

    NOTE: This is not commutative (merge(a,b) != merge(b,a)).
    """
    merged = {}

    left_keys = frozenset(left)
    right_keys = frozenset(right)

    # Items only in the left Mapping
    for key in left_keys - right_keys:
        merged[key] = left[key]

    # Items only in the right Mapping
    for key in right_keys - left_keys:
        merged[key] = right[key]

    # in both
    for key in left_keys & right_keys:
        left_value = left[key]
        right_value = right[key]

        # recursive merge
        if isinstance(left_value, Mapping) and isinstance(right_value, Mapping):
            merged[key] = merge(left_value, right_value)  # type: ignore
        # overwrite with right value
        else:
            merged[key] = right_value

    return merged


class Attr(Mapping[K, V], ABC):
    """
    A ``mixin`` class for a mapping that allows for attribute-style access of
    values.

    A key may be used as an attribute if:
     * It is a string
     * It matches ``^[A-Za-z][A-Za-z0-9_]*$`` (i.e., a public attribute)
     * The key doesn't overlap with any class attributes (for ``Attr``,
       those would be ``get``, ``items``, ``keys``, ``values``, ``mro``, and
       ``register``).

    If a value which is accessed as an attribute is a Sequence-type
    (and is not a string/bytes), it will be converted to a
    _sequence_type with any mappings within it converted to Attrs.

    NOTE:
        This means that if _sequence_type is not None, then a
        sequence accessed as an attribute will be a different object
        than if accessed as an attribute than if it is accessed as an
        item.
    """

    @abstractmethod
    def _configuration(self) -> Any:
        """All required state for building a new instance with the same
        settings as the current object."""

    @classmethod
    @abstractmethod
    def _constructor(cls, mapping: Mapping[K, V], configuration: Any) -> Attr[K, V]:
        """
        A standardized constructor used internally by Attr.

        mapping: A mapping of key-value pairs. It is HIGHLY recommended
            that you use this as the internal key-value pair mapping, as
            that will allow nested assignment (e.g., attr.foo.bar = baz)
        configuration: The return value of Attr._configuration
        """

    def __call__(self, key: K) -> Attr[K, V]:
        """
        Dynamically access a key-value pair.

        key: A key associated with a value in the mapping.

        This differs from __getitem__, because it returns a new instance
        of an Attr (if the value is a Mapping object).
        """
        if key not in self:
            raise AttributeError(
                f"'{self.__class__.__name__} instance has no attribute '{key}'"
            )

        return self._build(self[key])

    def __getattr__(self, key: Any) -> Any:
        """Access an item as an attribute."""
        if key not in self or not self._valid_name(key):
            raise AttributeError(
                f"'{self.__class__.__name__}' instance has no attribute '{key}'"
            )

        return self._build(self[key])

    def __add__(self, other: Mapping[K, V]) -> Attr[K, V]:
        """
        Add a mapping to this Attr, creating a new, merged Attr.

        other: A mapping.

        NOTE: Addition is not commutative. a + b != b + a.
        """
        if not isinstance(other, Mapping):
            return NotImplemented

        return self._constructor(merge(self, other), self._configuration())

    def __radd__(self, other: Mapping[K, V]) -> Attr[K, V]:
        """
        Add this Attr to a mapping, creating a new, merged Attr.

        other: A mapping.

        NOTE: Addition is not commutative. a + b != b + a.
        """
        if not isinstance(other, Mapping):
            return NotImplemented

        return self._constructor(merge(other, self), self._configuration())

    def _build(self, obj: Any) -> Attr[K, V]:
        """
        Conditionally convert an object to allow for recursive mapping access.

        obj: An object that was a key-value pair in the mapping. If obj
            is a mapping, self._constructor(obj, self._configuration())
            will be called. If obj is a non-string/bytes sequence, and
            self._sequence_type is not None, the obj will be converted
            to type _sequence_type and build will be called on its
            elements.
        """
        if isinstance(obj, Mapping):
            obj = self._constructor(obj, self._configuration())
        elif isinstance(obj, Sequence) and not isinstance(obj, (str, bytes)):
            sequence_type = getattr(self, "_sequence_type", None)
            if sequence_type:
                obj = sequence_type(self._build(element) for element in obj)

        return obj  # type: ignore

    @classmethod
    def _valid_name(cls, key: Any) -> bool:
        """
        Check whether a key is a valid attribute name.

        A key may be used as an attribute if:
         * It is a string
         * It matches /^[A-Za-z][A-Za-z0-9_]*$/ (i.e., a public attribute)
         * The key doesn't overlap with any class attributes (for Attr,
            those would be 'get', 'items', 'keys', 'values', 'mro', and
            'register').
        """
        return (
            isinstance(key, str)
            and not hasattr(cls, key)
            and VALID_KEY_RE.match(key) is not None
        )


class MutableAttr(Attr[str, V], MutableMapping[str, V], ABC):
    """A ``mixin`` mapping class that allows for attribute-style access of values."""

    def _setattr(self, key: str, value: Any) -> None:
        """Add an attribute to the object, without attempting to add it as a
        key to the mapping."""
        super().__setattr__(key, value)

    def __setattr__(self, key: str, value: V) -> None:
        """
        Add an attribute.

        key: The name of the attribute
        value: The attributes contents
        """
        if self._valid_name(key):
            self[key] = value
        elif getattr(self, "_allow_invalid_attributes", True):
            super().__setattr__(key, value)
        else:
            raise TypeError(
                f"'{self.__class__.__name__}' does not allow attribute creation."
            )

    def _delattr(self, key: str) -> None:
        """Delete an attribute from the object, without attempting to remove it
        from the mapping."""
        super().__delattr__(key)

    def __delattr__(self, key: str, force: bool = False) -> None:
        """
        Delete an attribute.

        key: The name of the attribute
        """
        if self._valid_name(key):
            del self[key]
        elif getattr(self, "_allow_invalid_attributes", True):
            super().__delattr__(key)
        else:
            raise TypeError(
                f"'{self.__class__.__name__}' does not allow attribute deletion."
            )


class AttrDict(Dict[str, V], MutableAttr[V]):
    """A dict that implements MutableAttr."""

    _sequence_type: type
    _allow_invalid_attributes: bool

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._setattr("_sequence_type", tuple)
        self._setattr("_allow_invalid_attributes", False)

    def _configuration(self) -> type:
        """The configuration for an attrmap instance."""
        return self._sequence_type

    def __getstate__(self) -> tuple[Mapping[str, V], type, bool]:
        """Serialize the object."""
        return self.copy(), self._sequence_type, self._allow_invalid_attributes

    def __setstate__(self, state: tuple[Mapping[str, V], type, bool]) -> None:
        """Deserialize the object."""
        mapping, sequence_type, allow_invalid_attributes = state
        self.update(mapping)
        self._setattr("_sequence_type", sequence_type)
        self._setattr("_allow_invalid_attributes", allow_invalid_attributes)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({super().__repr__()})"

    @classmethod
    def _constructor(
        cls,
        mapping: Mapping[str, V],
        configuration: type,
    ) -> AttrDict[V]:
        """A standardized constructor."""
        attr = cls(mapping)
        attr._setattr("_sequence_type", configuration)
        return attr



import sys
from contextlib import suppress
from time import sleep

import pytest
from qbittorrentapi import APINames
from qbittorrentapi._version_support import v
from qbittorrentapi.exceptions import APIError, Conflict409Error
from qbittorrentapi.rss import RSSitemsDictionary

from tests.utils import check, retry

FOLDER_ONE = "testFolderOne"
FOLDER_TWO = "testFolderTwo"

ITEM_ONE = "RSSOne"
ITEM_TWO = "RSSTwo"
RSS_NAME = "DistroWatch - Torrents"
RSS_URL = (
    "https://gist.githubusercontent.com/rmartin16/"
    "d615e1066f54186b44e8018d31af18f1/raw/b59cdc878fedfaf08efe6fc4321d18e8ded01e09/rss.xml"
)


@retry(3)
def delete_feed(client, name):
    with suppress(Conflict409Error):
        client.rss_remove_item(item_path=name)
        check(lambda: client.rss_items(), name, reverse=True, negate=True)


@pytest.fixture(scope="function", autouse=True)
def rss_feed(client, api_version):
    if v(api_version) >= v("2.2"):
        try:
            client.app.preferences = dict(rss_auto_downloading_enabled=False)
            # refreshing the feed is finicky...so try several times if necessary
            done = False
            for i in range(5):
                delete_feed(client, ITEM_ONE)
                delete_feed(client, RSS_NAME)
                client.rss.add_feed(url=RSS_URL, item_path=ITEM_ONE)
                check(lambda: client.rss_items(), ITEM_ONE, reverse=True)
                # wait until feed is refreshed
                for j in range(20):
                    if client.rss.items.with_data[ITEM_ONE]["articles"]:
                        done = True
                        yield ITEM_ONE
                        break
                    sleep(0.25)
                if done:
                    break
            else:
                raise Exception("RSS Feed '%s' did not refresh..." % ITEM_ONE)
        finally:
            delete_feed(client, ITEM_ONE)
            delete_feed(client, RSS_NAME)
    else:
        yield ""


@pytest.mark.skipif(sys.version_info < (3, 9), reason="removeprefix not in 3.8")
def test_methods(client):
    namespace = APINames.RSS
    all_dotted_methods = set(dir(getattr(client, namespace)))

    for meth in [meth for meth in dir(client) if meth.startswith(f"{namespace}_")]:
        assert meth.removeprefix(f"{namespace}_") in all_dotted_methods


@pytest.mark.skipif_before_api_version("2.2.1")
@pytest.mark.parametrize(
    "refresh_item_func",
    ["rss_refresh_item", "rss_refreshItem", "rss.refresh_item", "rss.refreshItem"],
)
def test_rss_refresh_item(client, rss_feed, refresh_item_func):
    last_log_id = client.log.main()[-1].id

    client.func(refresh_item_func)(item_path=rss_feed)

    check(
        lambda: [e.message for e in client.log.main(last_known_id=last_log_id)],
        "RSS feed at '%s' updated. Added 0 new articles." % RSS_URL,
        reverse=True,
    )


# inconsistent behavior with endpoint for API version 2.2
@pytest.mark.skipif_after_api_version("2.2")
@pytest.mark.parametrize(
    "refresh_item_func",
    ["rss_refresh_item", "rss_refreshItem", "rss.refresh_item", "rss.refreshItem"],
)
def test_rss_refresh_item_not_implemented(client, refresh_item_func):
    with pytest.raises(NotImplementedError):
        client.func(refresh_item_func)()


@pytest.mark.skipif_before_api_version("2.2")
@pytest.mark.parametrize("items_func", ["rss_items", "rss.items"])
def test_rss_items(client, rss_feed, items_func):
    check(lambda: client.func(items_func)(), rss_feed, reverse=True)
    check(
        lambda: client.func(items_func)(include_feed_data=True),
        rss_feed,
        reverse=True,
    )
    check(
        lambda: client.func(items_func)(include_feed_data=True)[rss_feed],
        "articles",
        reverse=True,
    )

    if "." in items_func:
        check(
            lambda: client.func(items_func).without_data,
            rss_feed,
            reverse=True,
        )
        check(
            lambda: client.func(items_func).with_data[rss_feed],
            "articles",
            reverse=True,
        )


@pytest.mark.skipif_before_api_version("2.2")
@pytest.mark.parametrize(
    "add_feed_func", ["rss_add_feed", "rss_addFeed", "rss.add_feed", "rss.addFeed"]
)
@pytest.mark.parametrize("feed_url, path", [(RSS_URL, "/path/my-feed"), (RSS_URL, "")])
def test_rss_add_feed(client, add_feed_func, feed_url, path):
    @retry(5)
    def run_test():
        delete_feed(client, ITEM_ONE)
        delete_feed(client, RSS_NAME)

        try:
            client.func(add_feed_func)(url=feed_url, item_path=path)
            check(lambda: client.rss_items(), path or feed_url, reverse=True)
        finally:
            delete_feed(client, path or feed_url)
            delete_feed(client, RSS_NAME)

    run_test()


@pytest.mark.skipif_before_api_version("2.9.1")
@pytest.mark.parametrize(
    "set_feed_func",
    ["rss_set_feed_url", "rss_setFeedURL", "rss.set_feed_url", "rss.setFeedURL"],
)
def test_rss_set_feed_url(client, rss_feed, set_feed_func):
    curr_feed_url = client.rss_items()[rss_feed].url
    new_feed_url = curr_feed_url + "asdf"
    client.func(set_feed_func)(url=new_feed_url, item_path=rss_feed)
    assert new_feed_url == client.rss_items()[rss_feed].url


@pytest.mark.skipif_after_api_version("2.9.1")
@pytest.mark.parametrize(
    "set_feed_func",
    ["rss_set_feed_url", "rss_setFeedURL", "rss.set_feed_url", "rss.setFeedURL"],
)
def test_rss_set_feed_url_not_implemented(client, set_feed_func):
    with pytest.raises(NotImplementedError):
        client.func(set_feed_func)()


@pytest.mark.skipif_before_api_version("2.2")
@pytest.mark.parametrize(
    "remove_item_func",
    ["rss_remove_item", "rss_removeItem", "rss.remove_item", "rss.removeItem"],
)
def test_rss_remove_feed(client, rss_feed, remove_item_func):
    client.func(remove_item_func)(item_path=rss_feed)
    check(lambda: client.rss_items(), rss_feed, reverse=True, negate=True)


@pytest.mark.parametrize(
    "add_folder_func, remove_item_func",
    [
        ("rss_add_folder", "rss_remove_item"),
        ("rss_addFolder", "rss_removeItem"),
        ("rss.add_folder", "rss.remove_item"),
        ("rss.addFolder", "rss.removeItem"),
    ],
)
def test_rss_add_remove_folder(client, add_folder_func, remove_item_func):
    name = "test_isos"

    client.func(add_folder_func)(folder_path=name)
    check(lambda: client.rss_items(), name, reverse=True)
    client.func(remove_item_func)(item_path=name)
    check(lambda: client.rss_items(), name, reverse=True, negate=True)


@pytest.mark.skipif_before_api_version("2.2")
@pytest.mark.parametrize(
    "move_func", ["rss_move_item", "rss_moveItem", "rss.move_item", "rss.moveItem"]
)
def test_rss_move(client, rss_feed, move_func):
    new_name = "new_loc"
    try:
        client.func(move_func)(orig_item_path=rss_feed, new_item_path=new_name)
        check(lambda: client.rss_items(), new_name, reverse=True)
    finally:
        with suppress(APIError):
            client.rss_remove_item(item_path=new_name)


@pytest.mark.skipif_before_api_version("2.5.1")
@pytest.mark.parametrize(
    "mark_read_func",
    ["rss_mark_as_read", "rss_markAsRead", "rss.mark_as_read", "rss.markAsRead"],
)
def test_rss_mark_as_read(client, rss_feed, mark_read_func):
    item_id = client.rss.items.with_data[rss_feed]["articles"][0]["id"]
    client.func(mark_read_func)(item_path=rss_feed, article_id=item_id)
    check(
        lambda: client.rss.items.with_data[rss_feed]["articles"][0],
        "isRead",
        reverse=True,
    )


@pytest.mark.skipif_after_api_version("2.5.1")
@pytest.mark.parametrize(
    "mark_read_func",
    ["rss_mark_as_read", "rss_markAsRead", "rss.mark_as_read", "rss.markAsRead"],
)
def test_rss_mark_as_read_not_implemented(client, mark_read_func):
    with pytest.raises(NotImplementedError):
        client.func(mark_read_func)()


@pytest.mark.skipif_before_api_version("2.2")
@pytest.mark.parametrize(
    "set_rule_func, rules_func, rename_rule_func, "
    "matching_func, remove_rule_func, remove_item_func",
    (
        (
            "rss_set_rule",
            "rss_rules",
            "rss_rename_rule",
            "rss_matching_articles",
            "rss_remove_rule",
            "rss_remove_item",
        ),
        (
            "rss_setRule",
            "rss_rules",
            "rss_renameRule",
            "rss_matchingArticles",
            "rss_removeRule",
            "rss_removeItem",
        ),
        (
            "rss.set_rule",
            "rss.rules",
            "rss.rename_rule",
            "rss.matching_articles",
            "rss.remove_rule",
            "rss.remove_item",
        ),
        (
            "rss.setRule",
            "rss.rules",
            "rss.renameRule",
            "rss.matchingArticles",
            "rss.removeRule",
            "rss.removeItem",
        ),
    ),
)
def test_rss_rules(
    client,
    api_version,
    set_rule_func,
    rules_func,
    rename_rule_func,
    matching_func,
    remove_rule_func,
    remove_item_func,
):
    def check_for_rule(name):
        try:
            client.func(rules_func)()
            check(lambda: client.func(rules_func)(), name, reverse=True)
        except TypeError:
            check(lambda: client.func(rules_func), name, reverse=True)

    rule_name = ITEM_ONE + "Rule"
    rule_name_new = rule_name + "New"
    rule_def = {"enabled": True, "affectedFeeds": RSS_URL, "addPaused": True}
    try:
        client.func(set_rule_func)(rule_name=rule_name, rule_def=rule_def)
        check_for_rule(rule_name)

        if v(api_version) >= v("2.6"):  # rename was broken for a bit
            client.func(rename_rule_func)(
                orig_rule_name=rule_name, new_rule_name=rule_name_new
            )
            check_for_rule(rule_name_new)
        if v(api_version) >= v("2.5.1"):
            assert isinstance(
                client.func(matching_func)(rule_name=rule_name), RSSitemsDictionary
            )
        else:
            with pytest.raises(NotImplementedError):
                client.func(matching_func)(rule_name=rule_name)
    finally:
        client.func(remove_rule_func)(rule_name=rule_name)
        client.func(remove_rule_func)(rule_name=rule_name_new)
        check(lambda: client.rss_rules(), rule_name, reverse=True, negate=True)
        client.func(remove_item_func)(item_path=ITEM_ONE)
        assert ITEM_TWO not in client.rss_items()
        check(lambda: client.rss_items(), ITEM_TWO, reverse=True, negate=True)


@pytest.mark.skipif_after_api_version("2.2")
@pytest.mark.parametrize(
    "matching_func",
    [
        "rss_matching_articles",
        "rss_matchingArticles",
        "rss.matching_articles",
        "rss.matchingArticles",
    ],
)
def test_rss_rules_not_implemented(client, matching_func):
    with pytest.raises(NotImplementedError):
        client.func(matching_func)()



from __future__ import annotations

from functools import wraps
from logging import Logger, getLogger
from types import TracebackType
from typing import TYPE_CHECKING

from requests import Response

from qbittorrentapi import Version
from qbittorrentapi.definitions import APIKwargsT, APINames, ClientCache
from qbittorrentapi.exceptions import LoginFailed, UnsupportedQbittorrentVersion
from qbittorrentapi.request import Request

if TYPE_CHECKING:
    from qbittorrentapi.client import Client

logger: Logger = getLogger(__name__)


class AuthAPIMixIn(Request):
    """
    Implementation of all ``Authorization`` API methods.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host="localhost:8080", username="admin", password="adminadmin")
        >>> _ = client.is_logged_in
        >>> client.auth_log_in(username="admin", password="adminadmin")
        >>> client.auth_log_out()
    """  # noqa: E501

    @property
    def auth(self) -> Authorization:
        """Allows for transparent interaction with Authorization endpoints."""
        if self._authorization is None:
            self._authorization = Authorization(client=self)
        return self._authorization

    authorization = auth

    @property
    def is_logged_in(self) -> bool:
        """
        Returns True if low-overhead API call succeeds; False otherwise.

        There isn't a reliable way to know if an existing session is still valid without
        attempting to use it. qBittorrent invalidates cookies when they expire.

        :returns: ``True``/``False`` if current authorization cookie is accepted by qBittorrent
        """  # noqa: E501
        try:
            # use _request_manager() directly so log in is not attempted
            self._request_manager(
                http_method="post",
                api_namespace=APINames.Application,
                api_method="version",
            )
        except Exception:
            return False
        else:
            return True

    def auth_log_in(
        self,
        username: str | None = None,
        password: str | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """
        Log in to qBittorrent host.

        :raises LoginFailed: if credentials failed to log in
        :raises Forbidden403Error: if user is banned...or not logged in

        :param username: username for qBittorrent client
        :param password: password for qBittorrent client
        """
        if username:
            self.username = username
            self._password = password or ""

        # trigger a (re-)initialization in case this is a new instance of qBittorrent
        self._initialize_context()

        creds = {"username": self.username, "password": self._password}
        auth_response = self._post_cast(
            _name=APINames.Authorization,
            _method="login",
            data=creds,
            response_class=Response,
            **kwargs,
        )

        if auth_response.text != "Ok.":
            logger.debug("Login failed")
            raise LoginFailed()
        logger.debug("Login successful")

        # check if the connected qBittorrent is fully supported by this Client yet
        if self._RAISE_UNSUPPORTEDVERSIONERROR:
            app_version = self.app_version()  # type: ignore
            api_version = self.app_web_api_version()  # type: ignore
            if not (
                Version.is_api_version_supported(api_version)
                and Version.is_app_version_supported(app_version)
            ):
                raise UnsupportedQbittorrentVersion(
                    "This version of qBittorrent is not fully supported => "
                    f"App Version: {app_version} API Version: {api_version}"
                )

    @property
    def _SID(self) -> str | None:
        """
        Authorization session cookie from qBittorrent using default cookie name `SID`.

        Backwards compatible for :meth:`~AuthAPIMixIn._session_cookie`.
        """
        return self._session_cookie()

    def _session_cookie(self, cookie_name: str = "SID") -> str | None:
        """
        Authorization session cookie from qBittorrent.

        :param cookie_name: Name of the authorization cookie; configurable after v4.5.0.
        """
        if self._http_session:
            return self._http_session.cookies.get(cookie_name, None)  # type: ignore
        return None

    def auth_log_out(self, **kwargs: APIKwargsT) -> None:
        """End session with qBittorrent."""
        self._post(_name=APINames.Authorization, _method="logout", **kwargs)

    def __enter__(self) -> Client:
        self.auth_log_in()
        return self  # type: ignore[return-value]

    def __exit__(
        self,
        exctype: type[BaseException] | None,
        excinst: BaseException | None,
        exctb: TracebackType | None,
    ) -> None:
        self.auth_log_out()


class Authorization(ClientCache[AuthAPIMixIn]):
    """
    Allows interaction with the ``Authorization`` API endpoints.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host="localhost:8080", username="admin", password="adminadmin")
        >>> is_logged_in = client.auth.is_logged_in
        >>> client.auth.log_in(username="admin", password="adminadmin")
        >>> client.auth.log_out()
    """  # noqa: E501

    @property
    @wraps(AuthAPIMixIn.is_logged_in)
    def is_logged_in(self) -> bool:
        return self._client.is_logged_in

    @wraps(AuthAPIMixIn.auth_log_in)
    def log_in(
        self,
        username: str | None = None,
        password: str | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        return self._client.auth_log_in(username=username, password=password, **kwargs)

    @wraps(AuthAPIMixIn.auth_log_out)
    def log_out(self, **kwargs: APIKwargsT) -> None:
        return self._client.auth_log_out(**kwargs)
